# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Processors and helpers specific to the :mod:`logging` module from the `Python
standard library <https://docs.python.org/>`_.

See also :doc:`structlog's standard library support <standard-library>`.
"""

from __future__ import annotations

import asyncio
import contextvars
import functools
import logging
import sys

from functools import partial
from typing import Any, Callable, Collection, Iterable, Sequence

from . import _config
from ._base import BoundLoggerBase
from ._frames import _find_first_app_frame_and_name, _format_stack
from ._log_levels import _LEVEL_TO_NAME, _NAME_TO_LEVEL, add_log_level
from .contextvars import merge_contextvars
from .exceptions import DropEvent
from .processors import StackInfoRenderer
from .typing import Context, EventDict, ExcInfo, Processor, WrappedLogger


__all__ = [
    "add_log_level_number",
    "add_log_level",
    "add_logger_name",
    "ExtraAdder",
    "BoundLogger",
    "filter_by_level",
    "get_logger",
    "LoggerFactory",
    "PositionalArgumentsFormatter",
    "ProcessorFormatter",
    "recreate_defaults",
    "render_to_log_kwargs",
]


def recreate_defaults(*, log_level: int | None = logging.NOTSET) -> None:
    """
    Recreate defaults on top of standard library's logging.

    The output looks the same, but goes through `logging`.

    As with vanilla defaults, the backwards-compatibility guarantees don't
    apply to the settings applied here.

    :param log_level: If `None`, don't configure standard library logging **at
        all**.

        Otherwise configure it to log to `sys.stdout` at *log_level*
        (``logging.NOTSET`` being the default).

        If you need more control over `logging`, pass `None` here and configure
        it yourself.

    .. versionadded:: 22.1
    """
    if log_level is not None:
        kw = {}
        # 3.7 doesn't have the force keyword and we don't care enough to
        # re-implement it.
        if sys.version_info >= (3, 8):
            kw = {"force": True}

        logging.basicConfig(
            format="%(message)s",
            stream=sys.stdout,
            level=log_level,
            **kw,  # type: ignore[arg-type]
        )

    _config.reset_defaults()
    _config.configure(
        processors=[
            merge_contextvars,
            add_log_level,
            StackInfoRenderer(),
            _config._BUILTIN_DEFAULT_PROCESSORS[-2],  # TimeStamper
            _config._BUILTIN_DEFAULT_PROCESSORS[-1],  # ConsoleRenderer
        ],
        wrapper_class=BoundLogger,
        logger_factory=LoggerFactory(),
    )


_SENTINEL = object()


class _FixedFindCallerLogger(logging.Logger):
    """
    Change the behavior of `logging.Logger.findCaller` to cope with
    *structlog*'s extra frames.
    """

    def findCaller(
        self, stack_info: bool = False, stacklevel: int = 1
    ) -> tuple[str, int, str, str | None]:
        """
        Finds the first caller frame outside of structlog so that the caller
        info is populated for wrapping stdlib.

        This logger gets set as the default one when using LoggerFactory.
        """
        sinfo: str | None
        f, name = _find_first_app_frame_and_name(["logging"])
        sinfo = _format_stack(f) if stack_info else None

        return f.f_code.co_filename, f.f_lineno, f.f_code.co_name, sinfo


class BoundLogger(BoundLoggerBase):
    """
    Python Standard Library version of `structlog.BoundLogger`.

    Works exactly like the generic one except that it takes advantage of
    knowing the logging methods in advance.

    Use it like::

        structlog.configure(
            wrapper_class=structlog.stdlib.BoundLogger,
        )

    It also contains a bunch of properties that pass-through to the wrapped
    `logging.Logger` which should make it work as a drop-in replacement.

    .. versionadded:: 23.1.0
       Async variants `alog()`, `adebug()`, `ainfo()`, and so forth.
    """

    _logger: logging.Logger

    def bind(self, **new_values: Any) -> BoundLogger:
        """
        Return a new logger with *new_values* added to the existing ones.
        """
        return super().bind(**new_values)  # type: ignore[return-value]

    def unbind(self, *keys: str) -> BoundLogger:
        """
        Return a new logger with *keys* removed from the context.

        :raises KeyError: If the key is not part of the context.
        """
        return super().unbind(*keys)  # type: ignore[return-value]

    def try_unbind(self, *keys: str) -> BoundLogger:
        """
        Like :meth:`unbind`, but best effort: missing keys are ignored.

        .. versionadded:: 18.2.0
        """
        return super().try_unbind(*keys)  # type: ignore[return-value]

    def new(self, **new_values: Any) -> BoundLogger:
        """
        Clear context and binds *initial_values* using `bind`.

        Only necessary with dict implementations that keep global state like
        those wrapped by `structlog.threadlocal.wrap_dict` when threads
        are re-used.
        """
        return super().new(**new_values)  # type: ignore[return-value]

    def debug(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """
        Process event and call `logging.Logger.debug` with the result.
        """
        return self._proxy_to_logger("debug", event, *args, **kw)

    def info(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """
        Process event and call `logging.Logger.info` with the result.
        """
        return self._proxy_to_logger("info", event, *args, **kw)

    def warning(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """
        Process event and call `logging.Logger.warning` with the result.
        """
        return self._proxy_to_logger("warning", event, *args, **kw)

    warn = warning

    def error(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """
        Process event and call `logging.Logger.error` with the result.
        """
        return self._proxy_to_logger("error", event, *args, **kw)

    def critical(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """
        Process event and call `logging.Logger.critical` with the result.
        """
        return self._proxy_to_logger("critical", event, *args, **kw)

    def exception(
        self, event: str | None = None, *args: Any, **kw: Any
    ) -> Any:
        """
        Process event and call `logging.Logger.error` with the result,
        after setting ``exc_info`` to `True`.
        """
        kw.setdefault("exc_info", True)

        return self.error(event, *args, **kw)

    def log(
        self, level: int, event: str | None = None, *args: Any, **kw: Any
    ) -> Any:
        """
        Process *event* and call the appropriate logging method depending on
        *level*.
        """
        return self._proxy_to_logger(_LEVEL_TO_NAME[level], event, *args, **kw)

    fatal = critical

    def _proxy_to_logger(
        self,
        method_name: str,
        event: str | None = None,
        *event_args: str,
        **event_kw: Any,
    ) -> Any:
        """
        Propagate a method call to the wrapped logger.

        This is the same as the superclass implementation, except that
        it also preserves positional arguments in the ``event_dict`` so
        that the stdlib's support for format strings can be used.
        """
        if event_args:
            event_kw["positional_args"] = event_args

        return super()._proxy_to_logger(method_name, event=event, **event_kw)

    # Pass-through attributes and methods to mimic the stdlib's logger
    # interface.

    @property
    def name(self) -> str:
        """
        Returns :attr:`logging.Logger.name`
        """
        return self._logger.name

    @property
    def level(self) -> int:
        """
        Returns :attr:`logging.Logger.level`
        """
        return self._logger.level

    @property
    def parent(self) -> Any:
        """
        Returns :attr:`logging.Logger.parent`
        """
        return self._logger.parent

    @property
    def propagate(self) -> bool:
        """
        Returns :attr:`logging.Logger.propagate`
        """
        return self._logger.propagate

    @property
    def handlers(self) -> Any:
        """
        Returns :attr:`logging.Logger.handlers`
        """
        return self._logger.handlers

    @property
    def disabled(self) -> int:
        """
        Returns :attr:`logging.Logger.disabled`
        """
        return self._logger.disabled

    def setLevel(self, level: int) -> None:
        """
        Calls :meth:`logging.Logger.setLevel` with unmodified arguments.
        """
        self._logger.setLevel(level)

    def findCaller(
        self, stack_info: bool = False
    ) -> tuple[str, int, str, str | None]:
        """
        Calls :meth:`logging.Logger.findCaller` with unmodified arguments.
        """
        return self._logger.findCaller(stack_info=stack_info)

    def makeRecord(
        self,
        name: str,
        level: int,
        fn: str,
        lno: int,
        msg: str,
        args: tuple[Any, ...],
        exc_info: ExcInfo,
        func: str | None = None,
        extra: Any = None,
    ) -> logging.LogRecord:
        """
        Calls :meth:`logging.Logger.makeRecord` with unmodified arguments.
        """
        return self._logger.makeRecord(
            name, level, fn, lno, msg, args, exc_info, func=func, extra=extra
        )

    def handle(self, record: logging.LogRecord) -> None:
        """
        Calls :meth:`logging.Logger.handle` with unmodified arguments.
        """
        self._logger.handle(record)

    def addHandler(self, hdlr: logging.Handler) -> None:
        """
        Calls :meth:`logging.Logger.addHandler` with unmodified arguments.
        """
        self._logger.addHandler(hdlr)

    def removeHandler(self, hdlr: logging.Handler) -> None:
        """
        Calls :meth:`logging.Logger.removeHandler` with unmodified arguments.
        """
        self._logger.removeHandler(hdlr)

    def hasHandlers(self) -> bool:
        """
        Calls :meth:`logging.Logger.hasHandlers` with unmodified arguments.

        Exists only in Python 3.
        """
        return self._logger.hasHandlers()

    def callHandlers(self, record: logging.LogRecord) -> None:
        """
        Calls :meth:`logging.Logger.callHandlers` with unmodified arguments.
        """
        self._logger.callHandlers(record)

    def getEffectiveLevel(self) -> int:
        """
        Calls :meth:`logging.Logger.getEffectiveLevel` with unmodified
        arguments.
        """
        return self._logger.getEffectiveLevel()

    def isEnabledFor(self, level: int) -> bool:
        """
        Calls :meth:`logging.Logger.isEnabledFor` with unmodified arguments.
        """
        return self._logger.isEnabledFor(level)

    def getChild(self, suffix: str) -> logging.Logger:
        """
        Calls :meth:`logging.Logger.getChild` with unmodified arguments.
        """
        return self._logger.getChild(suffix)

    # Non-Standard Async
    async def _dispatch_to_sync(
        self,
        meth: Callable[..., Any],
        event: str,
        args: tuple[Any, ...],
        kw: dict[str, Any],
    ) -> None:
        """
        Merge contextvars and log using the sync logger in a thread pool.
        """
        ctx = contextvars.copy_context()

        await asyncio.get_running_loop().run_in_executor(
            None,
            lambda: ctx.run(lambda: meth(event, *args, **kw)),
        )

    async def adebug(self, event: str, *args: Any, **kw: Any) -> None:
        """
        Log using `debug()`, but asynchronously in a separate thread.

        .. versionadded:: 23.1.0
        """
        await self._dispatch_to_sync(self.debug, event, args, kw)

    async def ainfo(self, event: str, *args: Any, **kw: Any) -> None:
        """
        Log using `info()`, but asynchronously in a separate thread.

        .. versionadded:: 23.1.0
        """
        await self._dispatch_to_sync(self.info, event, args, kw)

    async def awarning(self, event: str, *args: Any, **kw: Any) -> None:
        """
        Log using `warning()`, but asynchronously in a separate thread.

        .. versionadded:: 23.1.0
        """
        await self._dispatch_to_sync(self.warning, event, args, kw)

    async def aerror(self, event: str, *args: Any, **kw: Any) -> None:
        """
        Log using `error()`, but asynchronously in a separate thread.

        .. versionadded:: 23.1.0
        """
        await self._dispatch_to_sync(self.error, event, args, kw)

    async def acritical(self, event: str, *args: Any, **kw: Any) -> None:
        """
        Log using `critical()`, but asynchronously in a separate thread.

        .. versionadded:: 23.1.0
        """
        await self._dispatch_to_sync(self.critical, event, args, kw)

    afatal = acritical

    async def aexception(self, event: str, *args: Any, **kw: Any) -> None:
        """
        Log using `exception()`, but asynchronously in a separate thread.

        .. versionadded:: 23.1.0
        """
        # To make `log.exception("foo") work, we have to check if the user
        # passed an explicit exc_info and if not, supply our own.
        if kw.get("exc_info", True) is True and kw.get("exception") is None:
            kw["exc_info"] = sys.exc_info()

        await self._dispatch_to_sync(self.exception, event, args, kw)

    async def alog(
        self, level: Any, event: str, *args: Any, **kw: Any
    ) -> None:
        """
        Log using `log()`, but asynchronously in a separate thread.

        .. versionadded:: 23.1.0
        """
        await self._dispatch_to_sync(partial(self.log, level), event, args, kw)


def get_logger(*args: Any, **initial_values: Any) -> BoundLogger:
    """
    Only calls `structlog.get_logger`, but has the correct type hints.

    .. warning::

       Does **not** check whether -- or ensure that -- you've configured
       *structlog* for standard library :mod:`logging`!

       See :doc:`standard-library` for details.

    .. versionadded:: 20.2.0
    """
    return _config.get_logger(*args, **initial_values)


class AsyncBoundLogger:
    """
    Wraps a `BoundLogger` & exposes its logging methods as ``async`` versions.

    Instead of blocking the program, they are run asynchronously in a thread
    pool executor.

    This means more computational overhead per log call. But it also means that
    the processor chain (e.g. JSON serialization) and I/O won't block your
    whole application.

    Only available for Python 3.7 and later.

    :ivar structlog.stdlib.BoundLogger sync_bl: The wrapped synchronous logger.
       It is useful to be able to log synchronously occasionally.

    .. versionadded:: 20.2.0
    .. versionchanged:: 20.2.0 fix _dispatch_to_sync contextvars usage
    """

    __slots__ = ["sync_bl", "_loop"]

    sync_bl: BoundLogger

    # Blatant lie, we use a property for _context. Need this for Protocol
    # though.
    _context: Context

    _executor = None
    _bound_logger_factory = BoundLogger

    def __init__(
        self,
        logger: logging.Logger,
        processors: Iterable[Processor],
        context: Context,
        *,
        # Only as an optimization for binding!
        _sync_bl: Any = None,  # *vroom vroom* over purity.
        _loop: Any = None,
    ):
        if _sync_bl:
            self.sync_bl = _sync_bl
            self._loop = _loop

            return

        self.sync_bl = self._bound_logger_factory(
            logger=logger, processors=processors, context=context
        )
        self._loop = asyncio.get_running_loop()

    # We have to ignore the type because we've already declared it to ensure
    # we're a BindableLogger.
    # Instances would've been correctly recognized as such, however the class
    # not and we need the class in `structlog.configure()`.
    @property  # type: ignore[no-redef]
    def _context(self) -> Context:
        return self.sync_bl._context

    def bind(self, **new_values: Any) -> AsyncBoundLogger:
        return AsyncBoundLogger(
            # logger, processors and context are within sync_bl. These
            # arguments are ignored if _sync_bl is passed. *vroom vroom* over
            # purity.
            logger=None,  # type: ignore[arg-type]
            processors=(),
            context={},
            _sync_bl=self.sync_bl.bind(**new_values),
            _loop=self._loop,
        )

    def new(self, **new_values: Any) -> AsyncBoundLogger:
        return AsyncBoundLogger(
            # c.f. comment in bind
            logger=None,  # type: ignore[arg-type]
            processors=(),
            context={},
            _sync_bl=self.sync_bl.new(**new_values),
            _loop=self._loop,
        )

    def unbind(self, *keys: str) -> AsyncBoundLogger:
        return AsyncBoundLogger(
            # c.f. comment in bind
            logger=None,  # type: ignore[arg-type]
            processors=(),
            context={},
            _sync_bl=self.sync_bl.unbind(*keys),
            _loop=self._loop,
        )

    def try_unbind(self, *keys: str) -> AsyncBoundLogger:
        return AsyncBoundLogger(
            # c.f. comment in bind
            logger=None,  # type: ignore[arg-type]
            processors=(),
            context={},
            _sync_bl=self.sync_bl.try_unbind(*keys),
            _loop=self._loop,
        )

    async def _dispatch_to_sync(
        self,
        meth: Callable[..., Any],
        event: str,
        args: tuple[Any, ...],
        kw: dict[str, Any],
    ) -> None:
        """
        Merge contextvars and log using the sync logger in a thread pool.
        """
        ctx = contextvars.copy_context()

        await asyncio.get_running_loop().run_in_executor(
            self._executor,
            lambda: ctx.run(lambda: meth(event, *args, **kw)),
        )

    async def debug(self, event: str, *args: Any, **kw: Any) -> None:
        await self._dispatch_to_sync(self.sync_bl.debug, event, args, kw)

    async def info(self, event: str, *args: Any, **kw: Any) -> None:
        await self._dispatch_to_sync(self.sync_bl.info, event, args, kw)

    async def warning(self, event: str, *args: Any, **kw: Any) -> None:
        await self._dispatch_to_sync(self.sync_bl.warning, event, args, kw)

    warn = warning

    async def error(self, event: str, *args: Any, **kw: Any) -> None:
        await self._dispatch_to_sync(self.sync_bl.error, event, args, kw)

    async def critical(self, event: str, *args: Any, **kw: Any) -> None:
        await self._dispatch_to_sync(self.sync_bl.critical, event, args, kw)

    fatal = critical

    async def exception(self, event: str, *args: Any, **kw: Any) -> None:
        # To make `log.exception("foo") work, we have to check if the user
        # passed an explicit exc_info and if not, supply our own.
        ei = kw.pop("exc_info", None)
        if ei is None and kw.get("exception") is None:
            ei = sys.exc_info()

        kw["exc_info"] = ei

        await self._dispatch_to_sync(self.sync_bl.exception, event, args, kw)

    async def log(self, level: Any, event: str, *args: Any, **kw: Any) -> None:
        await self._dispatch_to_sync(
            partial(self.sync_bl.log, level), event, args, kw
        )


class LoggerFactory:
    """
    Build a standard library logger when an *instance* is called.

    Sets a custom logger using :func:`logging.setLoggerClass` so variables in
    log format are expanded properly.

    >>> from contrast_vendor.structlog import configure
    >>> from contrast_vendor.structlog.stdlib import LoggerFactory
    >>> configure(logger_factory=LoggerFactory())

    :param ignore_frame_names: When guessing the name of a logger, skip frames
        whose names *start* with one of these.  For example, in pyramid
        applications you'll want to set it to
        ``["venusian", "pyramid.config"]``. This argument is
        called *additional_ignores* in other APIs throughout *structlog*.
    """

    def __init__(self, ignore_frame_names: list[str] | None = None):
        self._ignore = ignore_frame_names
        logging.setLoggerClass(_FixedFindCallerLogger)

    def __call__(self, *args: Any) -> logging.Logger:
        """
        Deduce the caller's module name and create a stdlib logger.

        If an optional argument is passed, it will be used as the logger name
        instead of guesswork.  This optional argument would be passed from the
        :func:`structlog.get_logger` call.  For example
        ``structlog.get_logger("foo")`` would cause this method to be called
        with ``"foo"`` as its first positional argument.

        .. versionchanged:: 0.4.0
            Added support for optional positional arguments.  Using the first
            one for naming the constructed logger.
        """
        if args:
            return logging.getLogger(args[0])

        # We skip all frames that originate from within structlog or one of the
        # configured names.
        _, name = _find_first_app_frame_and_name(self._ignore)

        return logging.getLogger(name)


class PositionalArgumentsFormatter:
    """
    Apply stdlib-like string formatting to the ``event`` key.

    If the ``positional_args`` key in the event dict is set, it must
    contain a tuple that is used for formatting (using the ``%s`` string
    formatting operator) of the value from the ``event`` key.  This works
    in the same way as the stdlib handles arguments to the various log
    methods: if the tuple contains only a single `dict` argument it is
    used for keyword placeholders in the ``event`` string, otherwise it
    will be used for positional placeholders.

    ``positional_args`` is populated by `structlog.stdlib.BoundLogger` or
    can be set manually.

    The *remove_positional_args* flag can be set to `False` to keep the
    ``positional_args`` key in the event dict; by default it will be
    removed from the event dict after formatting a message.
    """

    def __init__(self, remove_positional_args: bool = True) -> None:
        self.remove_positional_args = remove_positional_args

    def __call__(
        self, _: WrappedLogger, __: str, event_dict: EventDict
    ) -> EventDict:
        args = event_dict.get("positional_args")

        # Mimic the formatting behaviour of the stdlib's logging module, which
        # accepts both positional arguments and a single dict argument. The
        # "single dict" check is the same one as the stdlib's logging module
        # performs in LogRecord.__init__().
        if args:
            if len(args) == 1 and isinstance(args[0], dict) and args[0]:
                args = args[0]

            event_dict["event"] %= args

        if self.remove_positional_args and args is not None:
            del event_dict["positional_args"]

        return event_dict


def filter_by_level(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """
    Check whether logging is configured to accept messages from this log level.

    Should be the first processor if stdlib's filtering by level is used so
    possibly expensive processors like exception formatters are avoided in the
    first place.

    >>> import logging
    >>> from contrast_vendor.structlog.stdlib import filter_by_level
    >>> logging.basicConfig(level=logging.WARN)
    >>> logger = logging.getLogger()
    >>> filter_by_level(logger, 'warn', {})
    {}
    >>> filter_by_level(logger, 'debug', {})
    Traceback (most recent call last):
    ...
    DropEvent
    """
    if logger.isEnabledFor(_NAME_TO_LEVEL[method_name]):
        return event_dict

    raise DropEvent


def add_log_level_number(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """
    Add the log level number to the event dict.

    Log level numbers map to the log level names. The Python stdlib uses them
    for filtering logic. This adds the same numbers so users can leverage
    similar filtering. Compare::

       level in ("warning", "error", "critical")
       level_number >= 30

    The mapping of names to numbers is in
    ``structlog.stdlib._log_levels._NAME_TO_LEVEL``.

    .. versionadded:: 18.2.0
    """
    event_dict["level_number"] = _NAME_TO_LEVEL[method_name]

    return event_dict


def add_logger_name(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """
    Add the logger name to the event dict.
    """
    record = event_dict.get("_record")
    if record is None:
        event_dict["logger"] = logger.name
    else:
        event_dict["logger"] = record.name
    return event_dict


_LOG_RECORD_KEYS = logging.LogRecord(
    "name", 0, "pathname", 0, "msg", (), None
).__dict__.keys()


class ExtraAdder:
    """
    Add extra attributes of `logging.LogRecord` objects to the event
    dictionary.

    This processor can be used for adding data passed in the ``extra``
    parameter of the `logging` module's log methods to the event dictionary.

    :param allow: An optional collection of attributes that, if present in
        `logging.LogRecord` objects, will be copied to event dictionaries.

        If ``allow`` is None all attributes of `logging.LogRecord` objects that
        do not exist on a standard `logging.LogRecord` object will be copied to
        event dictionaries.

    .. versionadded:: 21.5.0
    """

    __slots__ = ["_copier"]

    def __init__(self, allow: Collection[str] | None = None) -> None:
        self._copier: Callable[[EventDict, logging.LogRecord], None]
        if allow is not None:
            # The contents of allow is copied to a new list so that changes to
            # the list passed into the constructor does not change the
            # behaviour of this processor.
            self._copier = functools.partial(self._copy_allowed, [*allow])
        else:
            self._copier = self._copy_all

    def __call__(
        self, logger: logging.Logger, name: str, event_dict: EventDict
    ) -> EventDict:
        record: logging.LogRecord | None = event_dict.get("_record")
        if record is not None:
            self._copier(event_dict, record)
        return event_dict

    @classmethod
    def _copy_all(
        cls, event_dict: EventDict, record: logging.LogRecord
    ) -> None:
        for key, value in record.__dict__.items():
            if key not in _LOG_RECORD_KEYS:
                event_dict[key] = value

    @classmethod
    def _copy_allowed(
        cls,
        allow: Collection[str],
        event_dict: EventDict,
        record: logging.LogRecord,
    ) -> None:
        for key in allow:
            if key in record.__dict__:
                event_dict[key] = record.__dict__[key]


def render_to_log_kwargs(
    _: logging.Logger, __: str, event_dict: EventDict
) -> EventDict:
    """
    Render ``event_dict`` into keyword arguments for `logging.log`.

    The ``event`` field is translated into ``msg`` and the rest of the
    *event_dict* is added as ``extra``.

    This allows you to defer formatting to `logging`.

    .. versionadded:: 17.1.0
    .. versionchanged:: 22.1.0 ``exc_info``, ``stack_info``, and ``stackLevel``
       are passed as proper kwargs and not put into ``extra``.
    """
    return {
        "msg": event_dict.pop("event"),
        "extra": event_dict,
        **{
            kw: event_dict.pop(kw)
            for kw in ("exc_info", "stack_info", "stackLevel")
            if kw in event_dict
        },
    }


class ProcessorFormatter(logging.Formatter):
    r"""
    Call *structlog* processors on `logging.LogRecord`\s.

    This is an implementation of a `logging.Formatter` that can be used to
    format log entries from both *structlog* and `logging`.

    Its static method `wrap_for_formatter` must be the final processor in
    *structlog*'s processor chain.

    Please refer to :ref:`processor-formatter` for examples.

    :param foreign_pre_chain:
        If not `None`, it is used as a processor chain that is applied to
        **non**-*structlog* log entries before the event dictionary is passed
        to *processors*. (default: `None`)
    :param processors:
        A chain of *structlog* processors that is used to process **all** log
        entries. The last one must render to a `str` which then gets passed on
        to `logging` for output.

        Compared to *structlog*'s regular processor chains, there's a few
        differences:

        - The event dictionary contains two additional keys:

          #. ``_record``: a `logging.LogRecord` that either was created using
             `logging` APIs, **or** is a wrapped *structlog* log entry
             created by `wrap_for_formatter`.
          #. ``_from_structlog``: a `bool` that indicates whether or not
             ``_record`` was created by a *structlog* logger.

          Since you most likely don't want ``_record`` and
          ``_from_structlog`` in your log files,  we've added
          the static method `remove_processors_meta` to ``ProcessorFormatter``
          that you can add just before your renderer.

        - Since this is a `logging` *formatter*, raising `structlog.DropEvent`
          will crash your application.

    :param keep_exc_info: ``exc_info`` on `logging.LogRecord`\ s is
        added to the ``event_dict`` and removed afterwards. Set this to
        ``True`` to keep it on the `logging.LogRecord`. (default: False)
    :param keep_stack_info: Same as *keep_exc_info* except for ``stack_info``.
        (default: False)
    :param logger: Logger which we want to push through the *structlog*
        processor chain. This parameter is necessary for some of the
        processors like `filter_by_level`. (default: None)
    :param pass_foreign_args: If True, pass a foreign log record's
        ``args`` attribute to the ``event_dict`` under ``positional_args`` key.
        (default: False)
    :param processor:
        A single *structlog* processor used for rendering the event
        dictionary before passing it off to `logging`. Must return a `str`.
        The event dictionary does **not** contain ``_record`` and
        ``_from_structlog``.

        This parameter exists for historic reasons. Please consider using
        *processors* instead.

    :raises TypeError: If both or neither *processor* and *processors* are
        passed.

    .. versionadded:: 17.1.0
    .. versionadded:: 17.2.0 *keep_exc_info* and *keep_stack_info*
    .. versionadded:: 19.2.0 *logger*
    .. versionadded:: 19.2.0 *pass_foreign_args*
    .. versionadded:: 21.3.0 *processors*
    .. deprecated:: 21.3.0
       *processor* (singular) in favor of *processors* (plural). Removal is not
       planned.
    """

    def __init__(
        self,
        processor: Processor | None = None,
        processors: Sequence[Processor] | None = (),
        foreign_pre_chain: Sequence[Processor] | None = None,
        keep_exc_info: bool = False,
        keep_stack_info: bool = False,
        logger: logging.Logger | None = None,
        pass_foreign_args: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        fmt = kwargs.pop("fmt", "%(message)s")
        super().__init__(*args, fmt=fmt, **kwargs)  # type: ignore[misc]

        if processor and processors:
            raise TypeError(
                "The `processor` and `processors` arguments are mutually "
                "exclusive."
            )

        self.processors: Sequence[Processor]
        if processor is not None:
            self.processors = (self.remove_processors_meta, processor)
        elif processors:
            self.processors = processors
        else:
            raise TypeError(
                "Either `processor` or `processors` must be passed."
            )

        self.foreign_pre_chain = foreign_pre_chain
        self.keep_exc_info = keep_exc_info
        self.keep_stack_info = keep_stack_info
        self.logger = logger
        self.pass_foreign_args = pass_foreign_args

    def format(self, record: logging.LogRecord) -> str:
        """
        Extract *structlog*'s `event_dict` from ``record.msg`` and format it.

        *record* has been patched by `wrap_for_formatter` first though, so the
         type isn't quite right.
        """
        # Make a shallow copy of the record to let other handlers/formatters
        # process the original one
        record = logging.makeLogRecord(record.__dict__)

        logger = getattr(record, "_logger", _SENTINEL)
        meth_name = getattr(record, "_name", "__structlog_sentinel__")

        if logger is not _SENTINEL and meth_name != "__structlog_sentinel__":
            # Both attached by wrap_for_formatter
            if self.logger is not None:
                logger = self.logger
            meth_name = record._name  # type: ignore[attr-defined]

            # We need to copy because it's possible that the same record gets
            # processed by multiple logging formatters.  LogRecord.getMessage
            # would transform our dict into a str.
            ed = record.msg.copy()  # type: ignore[attr-defined]
            ed["_record"] = record
            ed["_from_structlog"] = True
        else:
            logger = self.logger
            meth_name = record.levelname.lower()
            ed = {
                "event": record.getMessage(),
                "_record": record,
                "_from_structlog": False,
            }

            if self.pass_foreign_args:
                ed["positional_args"] = record.args

            record.args = ()

            # Add stack-related attributes to event_dict and unset them
            # on the record copy so that the base implementation wouldn't
            # append stacktraces to the output.
            if record.exc_info:
                ed["exc_info"] = record.exc_info
            if record.stack_info:
                ed["stack_info"] = record.stack_info

            if not self.keep_exc_info:
                record.exc_text = None
                record.exc_info = None
            if not self.keep_stack_info:
                record.stack_info = None

            # Non-structlog allows to run through a chain to prepare it for the
            # final processor (e.g. adding timestamps and log levels).
            for proc in self.foreign_pre_chain or ():
                ed = proc(logger, meth_name, ed)

        for p in self.processors:
            ed = p(logger, meth_name, ed)

        record.msg = ed

        return super().format(record)

    @staticmethod
    def wrap_for_formatter(
        logger: logging.Logger, name: str, event_dict: EventDict
    ) -> tuple[tuple[EventDict], dict[str, dict[str, Any]]]:
        """
        Wrap *logger*, *name*, and *event_dict*.

        The result is later unpacked by `ProcessorFormatter` when
        formatting log entries.

        Use this static method as the renderer (i.e. final processor) if you
        want to use `ProcessorFormatter` in your `logging` configuration.
        """
        return (event_dict,), {"extra": {"_logger": logger, "_name": name}}

    @staticmethod
    def remove_processors_meta(
        _: WrappedLogger, __: str, event_dict: EventDict
    ) -> EventDict:
        """
        Remove ``_record`` and ``_from_structlog`` from *event_dict*.

        These keys are added to the event dictionary, before
        `ProcessorFormatter`'s *processors* are run.

        .. versionadded:: 21.3.0
        """
        del event_dict["_record"]
        del event_dict["_from_structlog"]

        return event_dict



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re
from contrast.agent.protect.rule.deserialization.custom_searcher import CustomSearcher
from contrast.utils.pattern_builder import PatternBuilder


class YAMLSearcher(CustomSearcher):
    ID = "UD-YAML-1"

    PYTHON_TAG = "!!python"

    def __init__(self):
        CustomSearcher.__init__(self, self.ID)

    @property
    def pattern(self):
        # !!(str|null|int|bool|float|binary|timestamp|omap|pairs|set|seq|map|python)

        return (
            PatternBuilder()
            .literally("!!")
            .any_of_these(
                [
                    "str",
                    "null",
                    "int",
                    "bool",
                    "float",
                    "binary",
                    "timestamp",
                    "omap",
                    "pairs",
                    "set",
                    "seq",
                    "map",
                    "python/name",
                    "python/module",
                    "python/object/new",
                    "python/object/apply",
                ]
            )
            .build()
        )

    def impact_of(self, value):
        count = 0
        impact = self.IMPACT_NONE

        matches = re.finditer(self.pattern, value)

        for match in matches:
            count += 1

            if match.group(0).startswith(self.PYTHON_TAG):
                return self.IMPACT_CRITICAL

        if count > 2:
            impact = self.IMPACT_MEDIUM
        elif count >= 1:
            impact = self.IMPACT_LOW

        return impact



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Rules for coverage:

- new routes on init get count 0
- new routes after init get count 1 per context
- routes without a method type specified on init are ['GET', 'POST']
- a route needs verb, route, url, and count
- url should be normalized uri
- route should be the path to the view function (aka controller for Java people)
- one method type per route

Example:
      GET /blog/foo/bar - app.blogs.find(request, ) 0
"""

import re


DEFAULT_ROUTE_METHODS = ("GET", "POST")


# A format template for constructing "path-component" regexps, ie between "/"-s in a URI
# This is used to build regexps for uri normalization
# We use a look-ahead to ensure correctness for back-to-back path components
_RE_PATH_COMPONENT_FORMAT = r"(^|/){regex}(?=(/|$))"

# examples:
# /123
# 123/
# 123
# /123/
# 123/123
RE_DIGITS_PATH_COMPONENT = _RE_PATH_COMPONENT_FORMAT.format(regex=r"\d+")

# example:
# /fd4b78312634a236d11da0f9c32526e5b8261afa
RE_HASH_PATH_COMPONENT = _RE_PATH_COMPONENT_FORMAT.format(
    regex=r"([a-fA-F0-9]{2}){16,}"
)

# example:
# /b09112a0-a58f-487a-ab4b-3608bd64fb3f/
RE_UUID_PATH_COMPONENT = _RE_PATH_COMPONENT_FORMAT.format(
    regex=r"[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}"
)

# Returns digits between '/' or end of string
FIND_REQEUST_PARAMS = r"(?<=\/)\d+(?=\/|$)"


def get_normalized_uri(path):
    """
    A best-effort to remove client-specific information from the path.

    Example:
    /user/123456/page/12 -> /user/{n}/page/{n}
    """
    result = path

    result = re.sub(RE_DIGITS_PATH_COMPONENT, r"\1{n}", result)
    result = re.sub(RE_HASH_PATH_COMPONENT, r"\1{hash}", result)
    result = re.sub(RE_UUID_PATH_COMPONENT, r"\1{uuid}", result)

    return result


def get_url_parameters(path):
    """
    A best-effort to remove client-specific information from the path.

    Example:
    /user/123456/page/12 -> /user/{n}/page/{n}
    """
    return re.findall(FIND_REQEUST_PARAMS, path)


def build_route(view_func_name, view_func):
    view_func_args = build_args_from_function(view_func)
    return view_func_name + view_func_args


def build_key(route_id, method):
    return "_".join([route_id, method])


def build_args_from_function(func):
    """
    Attempts to grab argument names from the function definition.

    Defaults to () if none exist
    If there is no view function, like in the case of a pure WSGI app, then the func will
    be a string like '/sqli' and we just return that.

    """
    method_arg_names = "()"
    if func is not None and hasattr(func, "__code__"):
        method_arg_names = str(
            func.__code__.co_varnames[0 : func.__code__.co_argcount]
        ).replace("'", "")
    elif isinstance(func, str):
        method_arg_names = func

    return method_arg_names



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes
from contrast.agent.policy.utils import CompositeNode


codecs_propagators = [
    CompositeNode(
        {
            "module": ["_codecs", "codecs"],
            "instance_method": True,
            "target": "RETURN",
        },
        [
            {
                "method_name": ["encode", "decode"],
                "source": "ARG_0,KWARG:obj",
                "action": "SPLAT",
            },
            {
                # No keyword arguments
                "method_name": [
                    "ascii_encode",
                    "charmap_encode",
                    "escape_encode",
                    "latin_1_encode",
                    "raw_unicode_escape_encode",
                    "readbuffer_encode",
                    "unicode_escape_encode",
                    "unicode_escape_decode",
                    "unicode_internal_encode",
                    "utf_16_encode",
                    "utf_16_be_encode",
                    "utf_16_le_encode",
                    "utf_32_encode",
                    "utf_32_be_encode",
                    "utf_32_le_encode",
                    "utf_7_encode",
                    "utf_8_encode",
                    "ascii_decode",
                    "charmap_decode",
                    "escape_decode",
                    "latin_1_decode",
                    "raw_unicode_escape_decode",
                    "unicode_internal_decode",
                    "utf_16_decode",
                    "utf_16_be_decode",
                    "utf_16_le_decode",
                    "utf_32_decode",
                    "utf_32_be_decode",
                    "utf_32_le_decode",
                    "utf_7_decode",
                    "utf_8_decode",
                ],
                "source": "ARG_0",
                "action": "CODECS_SPLAT",
            },
        ],
    ),
]


register_propagation_nodes(codecs_propagators)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class WsgiComplianceException(Exception):
    """
    Used to indicate that an application or server does not comply
    with the WSGI specification. See https://www.python.org/dev/peps/pep-0333
    """

    pass



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from io import open
import ast


def get_ast_parsed_code(path):
    """
    Open the path, retrieve text, and parse the
    text to create an ast tree of the code

    :param path: path to file location
    :return: ast representation of code in path.
    """
    with open(path, encoding="utf-8") as path_file:
        text = path_file.read()

    if isinstance(text, str):
        code_text = text.encode("utf-8", "ignore")
    else:
        code_text = str(text)

    parsed = ast.parse(code_text)

    return parsed



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from . import stream_propagator
from .base_propagator import BasePropagator, SUPPORTED_TYPES
from .append_propagator import AppendPropagator
from .center_propagator import CenterPropagator
from .codecs_splat_propagator import CodecsSplatPropagator
from .format_propagator import FormatPropagator
from .join_propagator import JoinPropagator
from .json_propagator import JsonPropagator
from .keep_propagator import KeepPropagator
from .prepend_propagator import PrependPropagator
from .replace_propagator import ReplacePropagator
from .safe_join_propagator import SafeJoinPropagator
from .safe_join_propagator_django import SafeJoinDjangoPropagator
from .slice_propagator import SlicePropagator
from .splat_propagator import SplatPropagator
from .split_propagator import SplitPropagator
from .starlette_safe_path_propagator import StarletteSafePathPropagator
from .reductive_propagator import ReductivePropagator
from .regex_propagator import RegexFindallPropagator, RegexSplitPropagator
from .db_write_propagator import DBWritePropagator
from .encode_html_splat_propagator import EncodeHtmlSplatPropagator
from .repr_propagator import ReprPropagator
from .tagger import Tagger


APPEND_ACTION = "APPEND"
CENTER_ACTION = "CENTER"
CODECS_SPLAT_ACTION = "CODECS_SPLAT"
FORMAT_ACTION = "FORMAT"
JOIN_ACTION = "JOIN"
JSON_ACTION = "JSON"
KEEP_ACTION = "KEEP"
PREPEND_ACTION = "PREPEND"
REGEX_FINDALL_ACTION = "REGEX_FINDALL"
REGEX_SPLIT_ACTION = "REGEX_SPLIT"
REPLACE_ACTION = "REPLACE"
REPR_ACTION = "REPR"
REMOVE_ACTION = "REMOVE"
SAFE_JOIN_ACTION = "SAFE_JOIN"
SAFE_JOIN_DJANGO_ACTION = "SAFE_JOIN_DJANGO"
SAFE_PATH_STARLETTE_ACTION = "STARLETTE_LOOKUP_PATH"
SLICE_ACTION = "SLICE"
SPLAT_ACTION = "SPLAT"
SPLIT_ACTION = "SPLIT"
TAG_ACTION = "TAG"
DB_WRITE_ACTION = "DB_WRITE"
ENCODE_HTML_SPLAT_ACTION = "ENCODE_HTML_SPLAT"
TAGGER_ACTION = "TAGGER"

# These propagators should not check the length of the target in order to determine
# whether to propagate.
IGNORE_LENGTH_ACTIONS = [DB_WRITE_ACTION, KEEP_ACTION, REPLACE_ACTION, SPLAT_ACTION]

PROPAGATOR_ACTIONS = {
    KEEP_ACTION: KeepPropagator,
    REPLACE_ACTION: ReplacePropagator,
    REMOVE_ACTION: ReductivePropagator,
    PREPEND_ACTION: PrependPropagator,
    APPEND_ACTION: AppendPropagator,
    CENTER_ACTION: CenterPropagator,
    SPLAT_ACTION: SplatPropagator,
    FORMAT_ACTION: FormatPropagator,
    SLICE_ACTION: SlicePropagator,
    JOIN_ACTION: JoinPropagator,
    SPLIT_ACTION: SplitPropagator,
    REGEX_FINDALL_ACTION: RegexFindallPropagator,
    REGEX_SPLIT_ACTION: RegexSplitPropagator,
    JSON_ACTION: JsonPropagator,
    ENCODE_HTML_SPLAT_ACTION: EncodeHtmlSplatPropagator,
    REPR_ACTION: ReprPropagator,
    DB_WRITE_ACTION: DBWritePropagator,
    SAFE_JOIN_ACTION: SafeJoinPropagator,
    SAFE_JOIN_DJANGO_ACTION: SafeJoinDjangoPropagator,
    SAFE_PATH_STARLETTE_ACTION: StarletteSafePathPropagator,
    CODECS_SPLAT_ACTION: CodecsSplatPropagator,
    TAGGER_ACTION: Tagger,
}

STREAM_ACTIONS = {
    "read": stream_propagator.propagate_stream_read,
    "read1": stream_propagator.propagate_stream_read,
    "readline": stream_propagator.propagate_stream_read,
    "readlines": stream_propagator.propagate_stream_readlines,
    "getvalue": stream_propagator.propagate_stream_getvalue,
}



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import re

from contrast.utils.string_utils import index_of_any


class ParseResults(object):
    def __init__(self, args, command_chains):
        self.args = args
        self.command_chains = command_chains


class CommandScanner(object):
    """
    Taken from Java Agent and Apache Commons
    """

    SINGLE_QUOTE = "'"
    DOUBLE_QUOTE = '"'

    CHAIN = ["&", ";", "|", ">", "<", "\n"]

    def __init__(self, executable):
        self.arguments = []
        self.command_chains = []
        self.executable = executable.replace("/", os.sep).replace("\\", os.sep)

    @staticmethod
    def parse(command):
        if not command:
            return None

        results = CommandScanner.translate_command_line(command)

        if results is None:
            return None

        try:
            command_scanner = CommandScanner(results.args[0])
        except Exception:
            return results

        for item in results.args[1:]:
            command_scanner.add_argument(item)

        command_scanner.command_chains = results.command_chains

        return results

    def add_argument(self, argument):
        if not argument:
            return

        self.arguments.append(self.quote_argument(argument))

    def quote_argument(self, argument):
        cleaned_argument = argument.strip()

        while cleaned_argument.startswith(
            self.SINGLE_QUOTE
        ) or cleaned_argument.startswith(self.DOUBLE_QUOTE):
            cleaned_argument = cleaned_argument[1:]

        while cleaned_argument.endswith(self.SINGLE_QUOTE) or cleaned_argument.endswith(
            self.DOUBLE_QUOTE
        ):
            cleaned_argument = cleaned_argument[:-1]

        if self.DOUBLE_QUOTE in cleaned_argument:
            if self.SINGLE_QUOTE in cleaned_argument:
                raise TypeError(
                    "Cannot handle single and double quotes in same argument"
                )

            result = ""
            result += self.SINGLE_QUOTE
            result += cleaned_argument
            result += self.SINGLE_QUOTE
            return result

        if " " in cleaned_argument:
            result = ""
            result += self.DOUBLE_QUOTE
            result += cleaned_argument
            result += self.DOUBLE_QUOTE
            return result

        return cleaned_argument

    @staticmethod
    def translate_command_line(command):
        normal = 0
        in_quote = 1
        in_double_quote = 2

        state = normal

        tokenized = re.split("([\"' ])", command)

        vectors = []
        chained_indices = []

        current = ""

        last_token_has_been_quoted = False

        index = 0

        for token in tokenized:
            index += len(token)

            if state == in_quote:
                if CommandScanner.SINGLE_QUOTE == token:
                    last_token_has_been_quoted = True
                    state = normal
                else:
                    current += token
            elif state == in_double_quote:
                if CommandScanner.DOUBLE_QUOTE == token:
                    last_token_has_been_quoted = True
                    state = normal
                else:
                    current += token
            else:
                if CommandScanner.SINGLE_QUOTE == token:
                    state = in_quote
                elif CommandScanner.DOUBLE_QUOTE == token:
                    state = in_double_quote
                elif token == " ":
                    if last_token_has_been_quoted or len(current) > -1:
                        vectors.append(current)
                        current = ""
                elif "#" in token:
                    break
                else:
                    current += token
                    chain_index = index_of_any(token, CommandScanner.CHAIN)

                    if chain_index != -1:
                        chained_indices.append(index - len(token) + chain_index)

                last_token_has_been_quoted = False

        if last_token_has_been_quoted or len(current) > -1:
            vectors.append(current)

        if state in (in_quote, in_double_quote):
            raise TypeError("Unbalanced quotes")

        return ParseResults(list(vectors), chained_indices)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from string import Formatter

from contrast.agent import scope
from contrast.agent.assess.adjusted_span import AdjustedSpan
from .base import StringToken, FormatToken as BaseFormatToken


class FormatToken(BaseFormatToken):
    def __init__(self, name, spec, conversion, offset, size, formatter):
        self.name = name
        self.spec = spec
        self.conversion = conversion
        self.offset = offset
        self.len = size
        self.formatter = formatter

    def format(self, args, kwargs):
        value = self.get_arg(args, kwargs)

        fmt = "{"
        if self.conversion:
            fmt += "!" + self.conversion
        if self.spec:
            fmt += ":" + self.spec
        fmt += "}"

        return value, fmt.format(value)

    def get_arg(self, args, kwargs):
        """
        Returns the value associated with this format token

        We need to convert the object if it is not actually a string type in order to
        determine whether/how to propagate. Unfortunately we only see the original
        object at the time the propagator is called, and so we're required to do the
        conversion manually. There is a small risk here if the __str__ or __repr__
        methods of the object are not idempotent for some reason, but that seems like
        a fairly extreme edge case.
        """
        field, _ = self.formatter.get_field(self.name, args, kwargs)
        if not isinstance(field, str) and not self.spec:
            # The default conversion behavior is to do nothing if no conversion is
            # given, but we need the default to be to call field.__str__, since that's
            # what format does by default. We pop out of scope in order to enable any
            # string conversion operations to propagate tags back to the field.
            with scope.pop_contrast_scope():
                return self.formatter.convert_field(field, self.conversion or "s")
        return field

    @property
    def span(self):
        return AdjustedSpan(self.offset, self.offset + self.len)

    def __len__(self):
        return self.len

    def __repr__(self):
        reprstr = "<{}({}, {}, {}, {}, len={})>"
        clsname = self.__class__.__name__
        return reprstr.format(
            clsname,
            repr(self.name),
            repr(self.spec),
            repr(self.conversion),
            self.offset,
            self.len,
        )


def tokenize_format(format_string):
    formatter = Formatter()

    tokens = []

    index = 0
    format_index = 0
    string = format_string

    nodes = list(formatter.parse(format_string))
    for literal, name, spec, conversion in nodes:
        # Handle case of '{{' or '}}'
        if literal and literal[-1] in ["{", "}"]:
            literal += literal[-1]

        offset = string.find(literal)
        string = string[offset:]

        if tokens and isinstance(tokens[-1], FormatToken):
            # Handle case of '{}}}'
            if literal and literal[-1] == "}":
                offset += 1
                string = string[1:]
            tokens[-1].len = offset

        index += offset

        if literal != "":
            tokens.append(StringToken(literal, index))
            index += len(literal)
            string = string[len(literal) :]

        if (name, spec, conversion) != (None, None, None):
            # Handle the case of adjacent format strings
            if tokens and isinstance(tokens[-1], FormatToken):
                # Find the offset immediately after the end of the last format
                offset = string.find("}") + 1
                tokens[-1].len = offset
                string = string[offset:]
                index += offset

            if not name:
                name = str(format_index)
                format_index += 1

            current_token = FormatToken(
                name, spec, conversion, index, offset, formatter
            )

            tokens.append(current_token)

    if isinstance(tokens[-1], FormatToken):
        current_token.len = len(format_string) - index

    return tokens



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys


def build_modules_to_ignore():
    """
    Get the set of module names that are
    1. in the modules to ignore list
    2. in sys.builtin_module_names
    3. NOT in the modules to allow list
    """
    return set(list(MODULES_TO_IGNORE + sys.builtin_module_names)).difference(
        MODULES_TO_ALLOW
    )


# this list contains modules to include despite their appearance in sys.builtin_module_names
MODULES_TO_ALLOW = ("__main__", "__builtin__", "builtins", "io")

MODULES_TO_IGNORE = (
    "__future__",
    "_ast",
    "_asyncio",
    "_bisect",
    "_blake2",
    "_bootlocale",
    "_bz2",
    "_collections",
    "_collections_abc",
    "_compat_pickle",
    "_compression",
    "_csv",
    "_ctypes",
    "_datetime",
    "_decimal",
    "_elementtree",
    "_frozen_importlib",
    "_frozen_importlib_external",
    "_functools",
    "_heapq",
    "_imp",
    "_jb_runner_tools",
    "_jb_serial_tree_manager",
    "_jb_utils",
    "_locale",
    "_lzma",
    "_markupbase",
    "_multiprocessing",
    "_opcode",
    "_operator",
    "_osx_support",
    "_pickle",
    "_posixsubprocess",
    "_random",
    "_sha3",
    "_signal",
    "_socket",
    "_sre",
    "_ssl",
    "_stat",
    "_string",
    "_struct",
    "_sysconfigdata_m_darwin_darwin",
    "_warnings",
    "_weakref",
    "_weakrefset",
    "_yaml",
    "abc",
    "argparse",
    "array",
    "ast",
    "atexit",
    "atomicwrites",
    "attr",
    "base64",
    "bdb",
    "bisect",
    "calendar",
    "cmd",
    "code",
    "codeop",
    "collections",
    "collections.abc",
    "concurrent",
    "configparser",
    "contextlib",
    "contrast",
    "copy",
    "copyreg",
    "coverage",
    "csv",
    "cython_runtime",
    "datetime",
    "decimal",
    "difflib",
    "dis",
    "doctest",
    "email",
    "email._encoded_words",
    "email._parseaddr",
    "email._policybase",
    "email.base64mime",
    "email.charset",
    "email.encoders",
    "email.errors",
    "email.feedparser",
    "email.header",
    "email.iterators",
    "email.message",
    "email.parser",
    "email.quoprimime",
    "email.utils",
    "encodings",
    "encodings.aliases",
    "enum",
    "errno",
    "fcntl",
    "fnmatch",
    "functools",
    "genericpath",
    "getopt",
    "gettext",
    "glob",
    "google",
    "grp",
    "gzip",
    "heapq",
    "hmac",
    "html",
    "html.entities",
    "html.parser",
    "imp",
    "importlib",
    "inspect",
    "ipaddress",
    "itertools",
    "keyword",
    "linecache",
    "locale",
    "logging",
    "lzma",
    "marshal",
    "math",
    "mimetypes",
    "mock",
    "mock.mock",
    "more_itertools",
    "more_itertools.more",
    "more_itertools.recipes",
    "ntpath",
    "numbers",
    "opcode",
    "operator",
    "optparse",
    "os",
    "os.path",
    "paste",
    "pbr",
    "pbr.version",
    "pdb",
    "pip",
    "pkgutil",
    "platform",
    "plistlib",
    "posixpath",
    "pprint",
    "psutil",
    "pwd",
    "py",
    "pyexpat",
    "pytz",
    "queue",
    "quopri",
    "reprlib",
    "select",
    "selectors",
    "shlex",
    "shutil",
    "signal",
    "site",
    "sitecustomize",
    "socket",
    "ssl",
    "stat",
    "string",
    "struct",
    "subprocess",
    "sys",
    "sysconfig",
    "tempfile",
    "termios",
    "textwrap",
    "time",
    "token",
    "tokenize",
    "traceback",
    "types",
    "unicodedata",
    "urllib.parse",
    "uu",
    "uuid",
    "warnings",
    "weakref",
    "wsgiref",
    "wsgiref.util",
    "zipfile",
    "zipimport",
    "zipp",
    "zlib",
    "zope",
)

ALL_MODULES_TO_IGNORE = build_modules_to_ignore()

# This list is carefully curated to ensure we don't exclude modules by accident.
# 'contrast' is not in this list because it's not
# unrealistic to think that a python library called contrastcolors or concurrent_stuff exists.
# Adding a dot ensures we only remove our own modules.
MODULES_TO_IGNORE_PREFIXES = (
    "_pytest",
    "attr.",
    "concurrent.",
    "contrast.",
    "contrast_vendor.",
    "coverage.",
    "ctypes",
    "distutils",
    "google.protobuf",
    "importlib.",
    "importlib_",
    "multiprocessing",
    "psutil.",
    "pkg_resources",
    "py.",
    "pyexpat.",
    "pluggy",
    "pytest",
    "pytz.",
    "sre_",
    "teamcity",
    "unittest",
    "wrapt",
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent.assess.rules.providers.hardcoded_value_rule import (
    HardcodedValueRule,
)
from contrast.utils.string_utils import ensure_string


class HardcodedPassword(HardcodedValueRule):
    """
    Determine if there are any passwords hardcoded into the sourcecode
    of the application. A constant is a password if:
    1) the name contains a PASSWORD_FIELD_NAMES value
    2) the name does not contain a NON_PASSWORD_PARTIAL_NAMES value
    3) the value is a String
    4) the value is not solely alphanumeric and '.' or '_' * note that mixing the
       characters counts as a violation of this rule
    """

    @property
    def name(self):
        return "hardcoded-password"

    # These are names, determined by the security team (Matt & Ar), that
    # indicate a field is likely to be a password or secret token of some
    # sort.
    PASSWORD_FIELD_NAMES = ["PASSWORD", "PASSKEY", "PASSPHRASE", "SECRET"]

    # These are markers whose presence indicates that a field is more
    # likely to be a descriptor or requirement than an actual password.
    # We should ignore fields that contain them.
    NON_PASSWORD_PARTIAL_NAMES = [
        "DATE",
        "FORGOT",
        "FORM",
        "ENCODE",
        "PATTERN",
        "PREFIX",
        "PROP",
        "SUFFIX",
        "URL",
        "BASE",
        "FILE",
    ]

    def is_name_valid(self, constant):
        not_in_password_fields = not any(
            self.fuzzy_match(constant, self.PASSWORD_FIELD_NAMES)
        )
        in_partial_names = any(
            self.fuzzy_match(constant, self.NON_PASSWORD_PARTIAL_NAMES)
        )
        return not_in_password_fields or in_partial_names

    def is_value_valid(self, value):
        return not self.probably_property_name(value)

    # If a field name matches an expected password field, we'll check it's
    # value to see if it looks like a placeholder. For our purposes,
    # placeholders will be any non-empty String conforming to the patterns
    # below.
    PROPERTY_NAME_PATTERN = re.compile("^[a-z]+[._][._a-z]*[a-z]+$")

    def probably_property_name(self, value):
        return re.match(self.PROPERTY_NAME_PATTERN, ensure_string(value)) is not None



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from collections import OrderedDict
import time
from weakref import WeakValueDictionary

from contrast.agent import scope
from contrast.assess_extensions import cs_str
from contrast.utils.decorators import fail_quietly
from contrast.utils.string_utils import truncated_signature

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class StringTracker(OrderedDict):
    AGE_OFF_THRESHOLD_SECS = 30

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pointer_to_props_map = WeakValueDictionary()

    def track(self, value):
        if not isinstance(value, (str, bytes, bytearray)):
            return None

        # ignore probable interned strings
        # TODO: PYT-2365 use is_interned here
        if len(value) < 2:
            return None

        if value and value not in self:
            from contrast.agent.assess.properties import Properties

            props = Properties(value)
            self.log_tracked(value)
            self[value] = props

        return self[value]

    def update_pointer_map(self, value, props):
        """
        Update mapping between str objects and their underlying char buffers

        We need to maintain this mapping in order to implement the hook for the
        exec statement in PY2. This is because the function we hook does not
        actually get a PyObject * as an argument, but only a char *. Since
        there is a 1-to-1 correspondence between the char * and the PyObject *,
        we can do a reverse lookup based on the pointer value in order to
        detect tracked strings that get passed to the exec statement.
        """
        ptr_val = cs_str.get_str_pointer(value)
        if ptr_val is None:
            return

        self.pointer_to_props_map[ptr_val] = props

    def lookup_by_pointer(self, ptr_val):
        return self.pointer_to_props_map.get(ptr_val)

    def log_tracked(self, value):
        self._truncate_and_log("tracking new string", value)

    def log_ageoff(self, value):
        self._truncate_and_log("aging off string from tracker", value)

    def _truncate_and_log(self, msg, value):
        with scope.contrast_scope():
            logger.debug("%s: %s", msg, truncated_signature(value))

    def update_properties(self, value, properties):
        self[value] = properties

    def __delitem__(self, key):
        return super().__delitem__(id(key))

    def __getitem__(self, key):
        return super().__getitem__(id(key))

    def __setitem__(self, key, value):
        if key in self:
            return super(OrderedDict, self).__setitem__(id(key), value)
        return super().__setitem__(id(key), value)

    def get(self, key, default=None):
        return super().get(id(key), default)

    if sys.version_info[:2] >= (3, 11):
        # Prior to 3.11, `pop` likely uses `__getitem__` under the hood. However, this
        # is no longer true in 3.11, which means we need to override this method.
        def pop(self, key):
            return super().pop(id(key))

    def __contains__(self, key):
        return super().__contains__(id(key))

    def get_by_id(self, key_id):
        return super().__getitem__(key_id)

    def values(self):
        "od.values() -> list of values in od"
        return [self.get_by_id(key) for key in self]

    def items(self):
        return [(key, self.get_by_id(key)) for key in self]

    def itervalues(self):
        "od.itervalues -> an iterator over the values in od"
        for k in self:
            yield self.get_by_id(k)

    def iteritems(self):
        "od.iteritems -> an iterator over the (key, value) pairs in od"
        for k in self:
            yield (k, self.get_by_id(k))

    @fail_quietly("Failure in string tracker ageoff")
    @scope.with_contrast_scope
    def ageoff(self):
        """
        Age-off string tracker entries older than predefined threshold.

        NOTE: It is only approximately true that this OrderedDict's values are sorted by
        timestamp. They are in insertion-order, but timestamps can change due to
        circumstances that are currently not well-understood. Do not assume timestamps
        are in order.

        TODO: PYT-2365 - we believe incorrect string tracker timestamps may be affected
        by string interning. When this ticket is completed, check to make sure ageoff
        is still working as expected.
        """
        now = time.time()
        logger.debug("String tracker length (pre-ageoff): %s", len(self))
        # Keys are IDs in this loop
        for key in list(self.keys()):
            props = super().__getitem__(key)
            if (now - props.timestamp) > self.AGE_OFF_THRESHOLD_SECS:
                self.log_ageoff(props.origin)
                try:
                    super().__delitem__(key)
                except KeyError:
                    # Handles potential race condition that may occur if
                    # another thread may have already deleted the key after
                    # this thread gets the key but before it deletes it.
                    # More likely to happen in multi-threaded apps and if
                    # and app receives many requests per second.
                    pass

        logger.debug("String tracker length (post-ageoff): %s", len(self))



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api import Route
from contrast.agent.middlewares.route_coverage.common import (
    build_route,
    get_normalized_uri,
    build_key,
)

from aiohttp.web_urldispatcher import DynamicResource


def create_aiohttp_routes(app):
    """
    Returns all the routes registered to a AioHttp app as a dict
    :param app: AioHttp app instance
    :return: dict {route_id:  api.Route}
    """
    routes = {}

    for app_route in app.router._resources:
        for resource in app_route._routes:
            view_func = resource.handler
            name = view_func.__name__
            route = build_route(name, view_func)

            route_id = str(id(view_func))

            _route_attr = (
                app_route._formatter
                if isinstance(app_route, DynamicResource)
                else app_route._path
            )
            method_type = resource.method
            key = build_key(route_id, method_type)
            routes[key] = Route(
                verb=method_type,
                url=get_normalized_uri(_route_attr),
                route=route,
            )

    return routes



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .django_httponly_rule import DjangoHttpOnlyRule
from .django_session_age_rule import DjangoSessionAgeRule
from .django_secure_flag_rule import DjangoSecureFlagRule
from .flask_httponly_rule import FlaskHttpOnlyRule
from .flask_session_age_rule import FlaskSessionAgeRule
from .flask_secure_flag_rule import FlaskSecureFlagRule



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators.base_propagator import (
    BasePropagator,
    SUPPORTED_TYPES,
)
from contrast.agent.assess.utils import (
    copy_from,
    is_tracked,
    get_properties,
    track_string,
)
from contrast.utils.assess.duck_utils import is_iterable, safe_iterator
from contrast.utils.decorators import fail_quietly

from contrast.utils.string_utils import ensure_binary


class SplatPropagator(BasePropagator):
    """
    Takes all tags from source string and applies each to entire length of the target
    """

    @fail_quietly("Splat propagation failed")
    def track_and_propagate(self, ret):
        """
        For splat, we'll track the target ourselves but ensure node tags and untags are applied
        """
        self.track_target()
        self.propagate()
        self.add_tags_and_properties(ret)
        self.reset_tags()

    def track_target(self):
        target_properties = get_properties(self.target)

        if not target_properties:
            target_properties = track_string(self.target)

        self.target_properties = target_properties

    def _propagate(self):
        if self._should_copy:
            track_string(self.sources[0])
            copy_from(self.target, self.sources[0], 0, self.node.untags)
            self.target_properties.cleanup_tags()
            return

        if not self.target or not isinstance(self.target, SUPPORTED_TYPES):
            return

        tracked_inputs = self._get_tracked_inputs()
        self._apply_splat_tags(tracked_inputs)

    @property
    def _should_copy(self):
        """
        Determine whether SPLAT operation should really be a copy (KEEP) instead

        If we can determine that the values of the source and target are the same, even
        if they are not the same type, then we should attempt to copy instead.  This
        will help improve accuracy. Specifically, it should help prevent some false
        negatives where a safe encoded tag may get splatted across an unsafe range as
        the result of an encode operation or some such.
        """
        if len(self.sources) != 1:
            return False

        try:
            if len(self.sources[0]) != len(self.target):
                return False

            return ensure_binary(self.sources[0]) == ensure_binary(self.target)
        except Exception:
            return False

    def _get_tracked_inputs(self):
        tracked_inputs = []

        def _append_if_tracked(val):
            if is_tracked(val):
                tracked_inputs.append(val)

        for source in self.sources:
            if isinstance(source, SUPPORTED_TYPES):
                _append_if_tracked(source)
            elif isinstance(source, dict):
                for key, value in source.items():
                    _append_if_tracked(key)
                    _append_if_tracked(value)
            elif is_iterable(source):
                for item in safe_iterator(source):
                    _append_if_tracked(item)

        return tracked_inputs

    def _apply_splat_tags(self, tracked_inputs):
        for item in tracked_inputs:
            if item is not self.target:
                item_properties = get_properties(item)

                if not item_properties:
                    continue

                for event in item_properties.events:
                    self.target_properties.events.append(event)

                self.splat_tags(item, self.target)

        self.target_properties.cleanup_tags()



import sys
from argparse import ArgumentParser

from . import hookspy


method_flags_map = {
    hookspy.METH_O: "METH_O",
    hookspy.METH_NOARGS: "METH_NOARGS",
    hookspy.METH_VARARGS: "METH_VARARGS",
    hookspy.METH_KEYWORDS: "METH_KEYWORDS",
    hookspy.METH_FASTCALL: "METH_FASTCALL",
    hookspy.METH_VARARGS | hookspy.METH_KEYWORDS: "METH_VARARGS|METH_KEYWORDS",
    hookspy.METH_FASTCALL | hookspy.METH_KEYWORDS: "METH_FASTCALL|METH_KEYWORDS",
    hookspy.METH_NEW: "METH_NEW",
    hookspy.METH_INIT: "METH_INIT",
}

type_spies = [
    ("unicode", "PyUnicode_Type", hookspy.find_unicode_hook),
    ("bytes", "PyBytes_Type", hookspy.find_bytes_hook),
    ("bytearray", "PyByteArray_Type", hookspy.find_bytearray_hook),
]


def main():
    p = ArgumentParser(description="find method hook locations for python string types")
    p.add_argument("method", type=str, help="Name of method")
    args = p.parse_args()

    print(sys.version)

    for name, typename, spy in type_spies:
        result = spy(args.method)
        if result is None:
            print('{} hook for "{}" not found'.format(name, args.method))
            continue

        index, flags = result
        flagname = method_flags_map.get(flags, "UNKNOWN")
        print("{:17} offset={:2}, flags={}".format(typename + ":", index, flagname))



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.config.base_django_config_rule import (
    BaseDjangoConfigRule,
)
from .secure_flag_rule import SecureFlagRuleMixin


class DjangoSecureFlagRule(SecureFlagRuleMixin, BaseDjangoConfigRule):
    SETTINGS_VALUE = "SESSION_COOKIE_SECURE"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.non_dataflow_rule import NonDataflowRule
from contrast.agent.policy.registry import register_trigger_rule

crypto_triggers = {
    "crypto-bad-mac": [
        {
            "module": "hashlib",
            "method_name": "new",
            "source": "ARG_0,KWARG:name",
            "good_value": "^(?:MDC2|RIPEMD160|SHA224|SHA256|SHA384|SHA512)",
        },
        {
            "module": "hashlib",
            "method_name": "md5",
        },
        {
            "module": "hashlib",
            "method_name": "sha1",
        },
        {
            "module": "Crypto.Hash.MD2",
            "class_name": "MD2Hash",
            "method_name": "__init__",
        },
        {
            "module": "Crypto.Hash.MD4",
            "class_name": "MD4Hash",
            "method_name": "__init__",
        },
        {
            "module": "Crypto.Hash.MD5",
            "class_name": "MD5Hash",
            "method_name": "__init__",
        },
        {
            "module": "Crypto.Hash.SHA1",
            "class_name": "SHA1Hash",
            "method_name": "__init__",
        },
        {
            "module": "Cryptodome.Hash.MD2",
            "class_name": "MD2Hash",
            "method_name": "__init__",
        },
        {
            "module": "Cryptodome.Hash.MD4",
            "class_name": "MD4Hash",
            "method_name": "__init__",
        },
        {
            "module": "Cryptodome.Hash.MD5",
            "class_name": "MD5Hash",
            "method_name": "__init__",
        },
        {
            "module": "Cryptodome.Hash.SHA1",
            "class_name": "SHA1Hash",
            "method_name": "__init__",
        },
    ],
    "crypto-weak-randomness": [
        {
            "module": "random",
            "method_name": "random",
        },
        {
            "module": "random",
            "method_name": "randint",
        },
        {
            "module": "random",
            "method_name": "randrange",
        },
        {
            "module": "random",
            "method_name": "uniform",
        },
    ],
    "crypto-bad-ciphers": [
        {
            "module": "Crypto.Cipher.Blowfish",
            "method_name": "new",
        },
        {
            "module": "Crypto.Cipher.DES",
            "method_name": "new",
        },
    ],
}


for name, nodes in crypto_triggers.items():
    register_trigger_rule(NonDataflowRule.from_nodes(name, nodes))



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import functools
from contrast_vendor.webob import Request

import contrast
from contrast.agent import agent_state, scope
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.agent.middlewares.environ_tracker import track_environ_sources
from contrast.utils.decorators import cached_property, fail_quietly
from contrast.utils.safe_import import safe_import_list
from contrast.utils.assess.duck_utils import safe_getattr_list

from contrast_vendor import structlog as logging
from contrast.utils import Profiler

logger = logging.getLogger("contrast")
DEFAULT_WSGI_NAME = "wsgi_app"


@functools.lru_cache(maxsize=1)
def _get_flask_types():
    return tuple(safe_import_list("flask.Flask"))


class WSGIMiddleware(BaseMiddleware):
    """
    Contrast middleware; PEP-333(3) WSGI-compliant
    """

    @scope.with_contrast_scope
    def __init__(
        self, wsgi_app, app_name=None, original_app=None, orig_pyramid_registry=None
    ):
        # We need to keep the `original_app` `orig_pyramid_registry` kwarg for now to prevent a breaking API
        # change
        del original_app
        del orig_pyramid_registry

        # TODO: PYT-2852 Revisit application name detection
        self.app_name = (
            app_name
            if app_name is not None
            else safe_getattr_list(
                wsgi_app,
                [
                    "__name__",
                    "name",
                ],
                DEFAULT_WSGI_NAME,
            )
        )

        super().__init__()

        if not agent_state.in_automatic_middleware() and isinstance(
            wsgi_app, _get_flask_types()
        ):
            # we need this to prevent a breaking API change when wrapping Flask apps
            wsgi_app = wsgi_app.wsgi_app

        self.wsgi_app = wsgi_app

    def __call__(self, environ, start_response):
        if contrast.CS__CONTEXT_TRACKER.current():
            # This can happen if a single app is wrapped by multiple instances of the
            # middleware (usually caused by automatic instrumentation)
            logger.debug("Detected preexisting request context")
            return self.wsgi_app(environ, start_response)

        self.request_path = environ.get("PATH_INFO", "")

        with Profiler(self.request_path):
            context = self.should_analyze_request(environ)
            if context:
                with contrast.CS__CONTEXT_TRACKER.lifespan(context):
                    return self.call_with_agent(context, environ, start_response)

            return self.call_without_agent(environ, start_response)

    def call_with_agent(self, context, environ, start_response):
        self.log_start_request_analysis()

        track_environ_sources("wsgi", context, environ)

        try:
            self.prefilter()

            response = Request(environ).get_response(self.wsgi_app)

            context.extract_response_to_context(response)

            self.postfilter(context)
            self.check_for_blocked(context)
            self.swap_environ_path(environ)  # should not be moved to finally

            return response(environ, start_response)

        finally:
            self.handle_ensure(context, context.request)
            self.log_end_request_analysis()
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()

    def swap_environ_path(self, environ):
        """
        See PYT-1742.

        Special behavior required for bottle+django to account for an unfortunate
        encoding behavior.

        In bottle, it occurs here:
        bottle.py:
        def _handle(self, environ):
            ...
            environ['PATH_INFO'] = path.encode('latin1').decode('utf8')
            ...

        This casing occurs after the application code is called, and will result in a
        `UnicodeEncodeError` when the agent attempts to access request.path.
        As a workaround, we store the original PATH_INFO - before bottle changed it
        during calling app code - and use it for agent purposes.
        """
        environ["PATH_INFO"] = self.request_path

    def call_without_agent(self, environ, start_response):
        """
        Normal without middleware call
        """
        super().call_without_agent()
        with scope.contrast_scope():
            return self.wsgi_app(environ, start_response)

    def get_route_coverage(self):
        """
        Unlike frameworks, WSGI is a spec, so it does not have a way to register routes.
        As such, there is no way to discover routes. Instead, routes will be added as
        they are visited via check for new routes.

        :return: empty dict
        """
        return {}

    @cached_property
    def trigger_node(self):
        """
        WSGI-specific trigger node used by reflected xss postfilter rule

        The rule itself is implemented in the base middleware but we need to
        provide a WSGI-specific trigger node for reporting purposes.
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.wsgi_app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    @fail_quietly("Unable to get WSGI view func")
    def get_view_func(self, _):
        """
        While most frameworks define view functions, WSGI doesn't so we will rely
        on the path information for reporting.
        If there is no path information, return an empty string

        :param request: RequestContext instance
        :return: string of path information for this request
        """
        return self.request_path or None

    @fail_quietly("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return url.replace("/", "")

    @cached_property
    def name(self):
        return "wsgi"



import re
import struct
import zlib
from base64 import b64encode
from datetime import datetime, timedelta
from hashlib import md5

from contrast_vendor.webob.byterange import ContentRange
from contrast_vendor.webob.cachecontrol import CacheControl, serialize_cache_control
from contrast_vendor.webob.compat import (
    PY2,
    bytes_,
    native_,
    string_types,
    text_type,
    url_quote,
    urlparse,
)
from contrast_vendor.webob.cookies import Cookie, make_cookie
from contrast_vendor.webob.datetime_utils import (
    parse_date_delta,
    serialize_date_delta,
    timedelta_to_seconds,
)
from contrast_vendor.webob.descriptors import (
    CHARSET_RE,
    SCHEME_RE,
    converter,
    date_header,
    header_getter,
    list_header,
    parse_auth,
    parse_content_range,
    parse_etag_response,
    parse_int,
    parse_int_safe,
    serialize_auth,
    serialize_content_range,
    serialize_etag_response,
    serialize_int,
)
from contrast_vendor.webob.headers import ResponseHeaders
from contrast_vendor.webob.request import BaseRequest
from contrast_vendor.webob.util import status_generic_reasons, status_reasons, warn_deprecation

try:
    import simplejson as json
except ImportError:
    import json

__all__ = ['Response']

_PARAM_RE = re.compile(r'([a-z0-9]+)=(?:"([^"]*)"|([a-z0-9_.-]*))', re.I)
_OK_PARAM_RE = re.compile(r'^[a-z0-9_.-]+$', re.I)

_gzip_header = b'\x1f\x8b\x08\x00\x00\x00\x00\x00\x02\xff'

_marker = object()

class Response(object):
    """
    Represents a WSGI response.

    If no arguments are passed, creates a :class:`~Response` that uses a
    variety of defaults. The defaults may be changed by sub-classing the
    :class:`~Response`. See the :ref:`sub-classing notes
    <response_subclassing_notes>`.

    :cvar ~Response.body: If ``body`` is a ``text_type``, then it will be
        encoded using either ``charset`` when provided or ``default_encoding``
        when ``charset`` is not provided if the ``content_type`` allows for a
        ``charset``. This argument is mutually  exclusive with ``app_iter``.

    :vartype ~Response.body: bytes or text_type

    :cvar ~Response.status: Either an :class:`int` or a string that is
        an integer followed by the status text. If it is an integer, it will be
        converted to a proper status that also includes the status text.  Any
        existing status text will be kept. Non-standard values are allowed.

    :vartype ~Response.status: int or str

    :cvar ~Response.headerlist: A list of HTTP headers for the response.

    :vartype ~Response.headerlist: list

    :cvar ~Response.app_iter: An iterator that is used as the body of the
        response. Should conform to the WSGI requirements and should provide
        bytes. This argument is mutually exclusive with ``body``.

    :vartype ~Response.app_iter: iterable

    :cvar ~Response.content_type: Sets the ``Content-Type`` header. If no
        ``content_type`` is provided, and there is no ``headerlist``, the
        ``default_content_type`` will be automatically set. If ``headerlist``
        is provided then this value is ignored.

    :vartype ~Response.content_type: str or None

    :cvar conditional_response: Used to change the behavior of the
        :class:`~Response` to check the original request for conditional
        response headers. See :meth:`~Response.conditional_response_app` for
        more information.

    :vartype conditional_response: bool

    :cvar ~Response.charset: Adds a ``charset`` ``Content-Type`` parameter. If
        no ``charset`` is provided and the ``Content-Type`` is text, then the
        ``default_charset`` will automatically be added.  Currently the only
        ``Content-Type``'s that allow for a ``charset`` are defined to be
        ``text/*``, ``application/xml``, and ``*/*+xml``. Any other
        ``Content-Type``'s will not have a ``charset`` added. If a
        ``headerlist`` is provided this value is ignored.

    :vartype ~Response.charset: str or None

    All other response attributes may be set on the response by providing them
    as keyword arguments. A :exc:`TypeError` will be raised for any unexpected
    keywords.

    .. _response_subclassing_notes:

    **Sub-classing notes:**

    * The ``default_content_type`` is used as the default for the
      ``Content-Type`` header that is returned on the response. It is
      ``text/html``.

    * The ``default_charset`` is used as the default character set to return on
      the ``Content-Type`` header, if the ``Content-Type`` allows for a
      ``charset`` parameter. Currently the only ``Content-Type``'s that allow
      for a ``charset`` are defined to be: ``text/*``, ``application/xml``, and
      ``*/*+xml``. Any other ``Content-Type``'s will not have a ``charset``
      added.

    * The ``unicode_errors`` is set to ``strict``, and access on a
      :attr:`~Response.text` will raise an error if it fails to decode the
      :attr:`~Response.body`.

    * ``default_conditional_response`` is set to ``False``. This flag may be
      set to ``True`` so that all ``Response`` objects will attempt to check
      the original request for conditional response headers. See
      :meth:`~Response.conditional_response_app` for more information.

    * ``default_body_encoding`` is set to 'UTF-8' by default. It exists to
      allow users to get/set the ``Response`` object using ``.text``, even if
      no ``charset`` has been set for the ``Content-Type``.
    """

    default_content_type = 'text/html'
    default_charset = 'UTF-8'
    unicode_errors = 'strict'
    default_conditional_response = False
    default_body_encoding = 'UTF-8'

    # These two are only around so that when people pass them into the
    # constructor they correctly get saved and set, however they are not used
    # by any part of the Response. See commit
    # 627593bbcd4ab52adc7ee569001cdda91c670d5d for rationale.
    request = None
    environ = None

    #
    # __init__, from_file, copy
    #

    def __init__(self, body=None, status=None, headerlist=None, app_iter=None,
                 content_type=None, conditional_response=None, charset=_marker,
                 **kw):
        # Do some sanity checking, and turn json_body into an actual body
        if app_iter is None and body is None and ('json_body' in kw or 'json' in kw):
            if 'json_body' in kw:
                json_body = kw.pop('json_body')
            else:
                json_body = kw.pop('json')
            body = json.dumps(json_body, separators=(',', ':')).encode('UTF-8')

            if content_type is None:
                content_type = 'application/json'

        if app_iter is None:
            if body is None:
                body = b''
        elif body is not None:
            raise TypeError(
                "You may only give one of the body and app_iter arguments")

        # Set up Response.status
        if status is None:
            self._status = '200 OK'
        else:
            self.status = status

        # Initialize headers
        self._headers = None
        if headerlist is None:
            self._headerlist = []
        else:
            self._headerlist = headerlist

        # Set the encoding for the Response to charset, so if a charset is
        # passed but the Content-Type does not allow for a charset, we can
        # still encode text_type body's.
        # r = Response(
        #   content_type='application/foo',
        #   charset='UTF-8',
        #   body=u'somebody')
        # Should work without issues, and the header will be correctly set to
        # Content-Type: application/foo with no charset on it.

        encoding = None
        if charset is not _marker:
            encoding = charset

        # Does the status code have a body or not?
        code_has_body = (
            self._status[0] != '1' and
            self._status[:3] not in ('204', '205', '304')
        )

        # We only set the content_type to the one passed to the constructor or
        # the default content type if there is none that exists AND there was
        # no headerlist passed. If a headerlist was provided then most likely
        # the ommission of the Content-Type is on purpose and we shouldn't try
        # to be smart about it.
        #
        # Also allow creation of a empty Response with just the status set to a
        # Response with empty body, such as Response(status='204 No Content')
        # without the default content_type being set (since empty bodies have
        # no Content-Type)
        #
        # Check if content_type is set because default_content_type could be
        # None, in which case there is no content_type, and thus we don't need
        # to anything

        content_type = content_type or self.default_content_type

        if headerlist is None and code_has_body and content_type:
            # Set up the charset, if the content_type doesn't already have one

            has_charset = 'charset=' in content_type

            # If the Content-Type already has a charset, we don't set the user
            # provided charset on the Content-Type, so we shouldn't use it as
            # the encoding for text_type based body's.
            if has_charset:
                encoding = None

            # Do not use the default_charset for the encoding because we
            # want things like
            # Response(content_type='image/jpeg',body=u'foo') to raise when
            # trying to encode the body.

            new_charset = encoding

            if (
                not has_charset and
                charset is _marker and
                self.default_charset
            ):
                new_charset = self.default_charset

            # Optimize for the default_content_type as shipped by
            # WebOb, becuase we know that 'text/html' has a charset,
            # otherwise add a charset if the content_type has a charset.
            #
            # Even if the user supplied charset explicitly, we do not add
            # it to the Content-Type unless it has has a charset, instead
            # the user supplied charset is solely used for encoding the
            # body if it is a text_type

            if (
                new_charset and
                (
                    content_type == 'text/html' or
                    _content_type_has_charset(content_type)
                )
            ):
                content_type += '; charset=' + new_charset

            self._headerlist.append(('Content-Type', content_type))

        # Set up conditional response
        if conditional_response is None:
            self.conditional_response = self.default_conditional_response
        else:
            self.conditional_response = bool(conditional_response)

        # Set up app_iter if the HTTP Status code has a body
        if app_iter is None and code_has_body:
            if isinstance(body, text_type):
                # Fall back to trying self.charset if encoding is not set. In
                # most cases encoding will be set to the default value.
                encoding = encoding or self.charset
                if encoding is None:
                    raise TypeError(
                        "You cannot set the body to a text value without a "
                        "charset")
                body = body.encode(encoding)
            app_iter = [body]

            if headerlist is not None:
                self._headerlist[:] = [
                    (k, v)
                    for (k, v)
                    in self._headerlist
                    if k.lower() != 'content-length'
                ]
            self._headerlist.append(('Content-Length', str(len(body))))
        elif app_iter is None and not code_has_body:
            app_iter = [b'']

        self._app_iter = app_iter

        # Loop through all the remaining keyword arguments
        for name, value in kw.items():
            if not hasattr(self.__class__, name):
                # Not a basic attribute
                raise TypeError(
                    "Unexpected keyword: %s=%r" % (name, value))
            setattr(self, name, value)

    @classmethod
    def from_file(cls, fp):
        """Reads a response from a file-like object (it must implement
        ``.read(size)`` and ``.readline()``).

        It will read up to the end of the response, not the end of the
        file.

        This reads the response as represented by ``str(resp)``; it
        may not read every valid HTTP response properly.  Responses
        must have a ``Content-Length``."""
        headerlist = []
        status = fp.readline().strip()
        is_text = isinstance(status, text_type)

        if is_text:
            _colon = ':'
            _http = 'HTTP/'
        else:
            _colon = b':'
            _http = b'HTTP/'

        if status.startswith(_http):
            (http_ver, status_num, status_text) = status.split(None, 2)
            status = '%s %s' % (native_(status_num), native_(status_text))

        while 1:
            line = fp.readline().strip()
            if not line:
                # end of headers
                break
            try:
                header_name, value = line.split(_colon, 1)
            except ValueError:
                raise ValueError('Bad header line: %r' % line)
            value = value.strip()
            headerlist.append((
                native_(header_name, 'latin-1'),
                native_(value, 'latin-1')
            ))
        r = cls(
            status=status,
            headerlist=headerlist,
            app_iter=(),
        )
        body = fp.read(r.content_length or 0)
        if is_text:
            r.text = body
        else:
            r.body = body
        return r

    def copy(self):
        """Makes a copy of the response."""
        # we need to do this for app_iter to be reusable
        app_iter = list(self._app_iter)
        iter_close(self._app_iter)
        # and this to make sure app_iter instances are different
        self._app_iter = list(app_iter)
        return self.__class__(
            status=self._status,
            headerlist=self._headerlist[:],
            app_iter=app_iter,
            conditional_response=self.conditional_response)

    #
    # __repr__, __str__
    #

    def __repr__(self):
        return '<%s at 0x%x %s>' % (self.__class__.__name__, abs(id(self)),
                                    self.status)

    def __str__(self, skip_body=False):
        parts = [self.status]
        if not skip_body:
            # Force enumeration of the body (to set content-length)
            self.body
        parts += map('%s: %s'.__mod__, self.headerlist)
        if not skip_body and self.body:
            parts += ['', self.body if PY2 else self.text]
        return '\r\n'.join(parts)

    #
    # status, status_code/status_int
    #

    def _status__get(self):
        """
        The status string.
        """
        return self._status

    def _status__set(self, value):
        try:
            code = int(value)
        except (ValueError, TypeError):
            pass
        else:
            self.status_code = code
            return
        if not PY2:
            if isinstance(value, bytes):
                value = value.decode('ascii')
        elif isinstance(value, text_type):
            value = value.encode('ascii')
        if not isinstance(value, str):
            raise TypeError(
                "You must set status to a string or integer (not %s)"
                % type(value))

        # Attempt to get the status code itself, if this fails we should fail
        try:
            # We don't need this value anywhere, we just want to validate it's
            # an integer. So we are using the side-effect of int() raises a
            # ValueError as a test
            int(value.split()[0])
        except ValueError:
            raise ValueError('Invalid status code, integer required.')
        self._status = value

    status = property(_status__get, _status__set, doc=_status__get.__doc__)

    def _status_code__get(self):
        """
        The status as an integer.
        """
        return int(self._status.split()[0])

    def _status_code__set(self, code):
        try:
            self._status = '%d %s' % (code, status_reasons[code])
        except KeyError:
            self._status = '%d %s' % (code, status_generic_reasons[code // 100])

    status_code = status_int = property(_status_code__get, _status_code__set,
                                        doc=_status_code__get.__doc__)

    #
    # headerslist, headers
    #

    def _headerlist__get(self):
        """
        The list of response headers.
        """
        return self._headerlist

    def _headerlist__set(self, value):
        self._headers = None
        if not isinstance(value, list):
            if hasattr(value, 'items'):
                value = value.items()
            value = list(value)
        self._headerlist = value

    def _headerlist__del(self):
        self.headerlist = []

    headerlist = property(_headerlist__get, _headerlist__set,
                          _headerlist__del, doc=_headerlist__get.__doc__)

    def _headers__get(self):
        """
        The headers in a dictionary-like object.
        """
        if self._headers is None:
            self._headers = ResponseHeaders.view_list(self._headerlist)
        return self._headers

    def _headers__set(self, value):
        if hasattr(value, 'items'):
            value = value.items()
        self.headerlist = value
        self._headers = None

    headers = property(_headers__get, _headers__set, doc=_headers__get.__doc__)

    #
    # body
    #

    def _body__get(self):
        """
        The body of the response, as a :class:`bytes`.  This will read in
        the entire app_iter if necessary.
        """
        app_iter = self._app_iter
#         try:
#             if len(app_iter) == 1:
#                 return app_iter[0]
#         except:
#             pass
        if isinstance(app_iter, list) and len(app_iter) == 1:
            return app_iter[0]
        if app_iter is None:
            raise AttributeError("No body has been set")
        try:
            body = b''.join(app_iter)
        finally:
            iter_close(app_iter)
        if isinstance(body, text_type):
            raise _error_unicode_in_app_iter(app_iter, body)
        self._app_iter = [body]
        if len(body) == 0:
            # if body-length is zero, we assume it's a HEAD response and
            # leave content_length alone
            pass
        elif self.content_length is None:
            self.content_length = len(body)
        elif self.content_length != len(body):
            raise AssertionError(
                "Content-Length is different from actual app_iter length "
                "(%r!=%r)"
                % (self.content_length, len(body))
            )
        return body

    def _body__set(self, value=b''):
        if not isinstance(value, bytes):
            if isinstance(value, text_type):
                msg = ("You cannot set Response.body to a text object "
                       "(use Response.text)")
            else:
                msg = ("You can only set the body to a binary type (not %s)" %
                       type(value))
            raise TypeError(msg)
        if self._app_iter is not None:
            self.content_md5 = None
        self._app_iter = [value]
        self.content_length = len(value)

#     def _body__del(self):
#         self.body = ''
#         #self.content_length = None

    body = property(_body__get, _body__set, _body__set)

    def _json_body__get(self):
        """
        Set/get the body of the response as JSON.

        .. note::

           This will automatically :meth:`~bytes.decode` the
           :attr:`~Response.body` as ``UTF-8`` on get, and
           :meth:`~str.encode` the :meth:`json.dumps` as ``UTF-8``
           before assigning to :attr:`~Response.body`.

        """
        # Note: UTF-8 is a content-type specific default for JSON
        return json.loads(self.body.decode('UTF-8'))

    def _json_body__set(self, value):
        self.body = json.dumps(value, separators=(',', ':')).encode('UTF-8')

    def _json_body__del(self):
        del self.body

    json = json_body = property(_json_body__get, _json_body__set, _json_body__del)

    def _has_body__get(self):
        """
        Determine if the the response has a :attr:`~Response.body`. In
        contrast to simply accessing :attr:`~Response.body`, this method
        will **not** read the underlying :attr:`~Response.app_iter`.
        """

        app_iter = self._app_iter

        if isinstance(app_iter, list) and len(app_iter) == 1:
            if app_iter[0] != b'':
                return True
            else:
                return False

        if app_iter is None: # pragma: no cover
            return False

        return True

    has_body = property(_has_body__get)

    #
    # text, unicode_body, ubody
    #

    def _text__get(self):
        """
        Get/set the text value of the body using the ``charset`` of the
        ``Content-Type`` or the ``default_body_encoding``.
        """
        if not self.charset and not self.default_body_encoding:
            raise AttributeError(
                "You cannot access Response.text unless charset or default_body_encoding"
                " is set"
            )
        decoding = self.charset or self.default_body_encoding
        body = self.body
        return body.decode(decoding, self.unicode_errors)

    def _text__set(self, value):
        if not self.charset and not self.default_body_encoding:
            raise AttributeError(
                "You cannot access Response.text unless charset or default_body_encoding"
                " is set"
            )
        if not isinstance(value, text_type):
            raise TypeError(
                "You can only set Response.text to a unicode string "
                "(not %s)" % type(value))
        encoding = self.charset or self.default_body_encoding
        self.body = value.encode(encoding)

    def _text__del(self):
        del self.body

    text = property(_text__get, _text__set, _text__del, doc=_text__get.__doc__)

    unicode_body = ubody = property(_text__get, _text__set, _text__del,
                                    "Deprecated alias for .text")

    #
    # body_file, write(text)
    #

    def _body_file__get(self):
        """
        A file-like object that can be used to write to the
        body.  If you passed in a list ``app_iter``, that ``app_iter`` will be
        modified by writes.
        """
        return ResponseBodyFile(self)

    def _body_file__set(self, file):
        self.app_iter = iter_file(file)

    def _body_file__del(self):
        del self.body

    body_file = property(_body_file__get, _body_file__set, _body_file__del,
                         doc=_body_file__get.__doc__)

    def write(self, text):
        if not isinstance(text, bytes):
            if not isinstance(text, text_type):
                msg = "You can only write str to a Response.body_file, not %s"
                raise TypeError(msg % type(text))
            if not self.charset:
                msg = ("You can only write text to Response if charset has "
                       "been set")
                raise TypeError(msg)
            text = text.encode(self.charset)
        app_iter = self._app_iter
        if not isinstance(app_iter, list):
            try:
                new_app_iter = self._app_iter = list(app_iter)
            finally:
                iter_close(app_iter)
            app_iter = new_app_iter
            self.content_length = sum(len(chunk) for chunk in app_iter)
        app_iter.append(text)
        if self.content_length is not None:
            self.content_length += len(text)

    #
    # app_iter
    #

    def _app_iter__get(self):
        """
        Returns the ``app_iter`` of the response.

        If ``body`` was set, this will create an ``app_iter`` from that
        ``body`` (a single-item list).
        """
        return self._app_iter

    def _app_iter__set(self, value):
        if self._app_iter is not None:
            # Undo the automatically-set content-length
            self.content_length = None
        self._app_iter = value

    def _app_iter__del(self):
        self._app_iter = []
        self.content_length = None

    app_iter = property(_app_iter__get, _app_iter__set, _app_iter__del,
                        doc=_app_iter__get.__doc__)

    #
    # headers attrs
    #

    allow = list_header('Allow', '14.7')
    # TODO: (maybe) support response.vary += 'something'
    # TODO: same thing for all listy headers
    vary = list_header('Vary', '14.44')

    content_length = converter(
        header_getter('Content-Length', '14.17'),
        parse_int, serialize_int, 'int')

    content_encoding = header_getter('Content-Encoding', '14.11')
    content_language = list_header('Content-Language', '14.12')
    content_location = header_getter('Content-Location', '14.14')
    content_md5 = header_getter('Content-MD5', '14.14')
    content_disposition = header_getter('Content-Disposition', '19.5.1')

    accept_ranges = header_getter('Accept-Ranges', '14.5')
    content_range = converter(
        header_getter('Content-Range', '14.16'),
        parse_content_range, serialize_content_range, 'ContentRange object')

    date = date_header('Date', '14.18')
    expires = date_header('Expires', '14.21')
    last_modified = date_header('Last-Modified', '14.29')

    _etag_raw = header_getter('ETag', '14.19')
    etag = converter(
        _etag_raw,
        parse_etag_response, serialize_etag_response,
        'Entity tag'
    )
    @property
    def etag_strong(self):
        return parse_etag_response(self._etag_raw, strong=True)

    location = header_getter('Location', '14.30')
    pragma = header_getter('Pragma', '14.32')
    age = converter(
        header_getter('Age', '14.6'),
        parse_int_safe, serialize_int, 'int')

    retry_after = converter(
        header_getter('Retry-After', '14.37'),
        parse_date_delta, serialize_date_delta, 'HTTP date or delta seconds')

    server = header_getter('Server', '14.38')

    # TODO: the standard allows this to be a list of challenges
    www_authenticate = converter(
        header_getter('WWW-Authenticate', '14.47'),
        parse_auth, serialize_auth,
    )

    #
    # charset
    #

    def _charset__get(self):
        """
        Get/set the ``charset`` specified in ``Content-Type``.

        There is no checking to validate that a ``content_type`` actually
        allows for a ``charset`` parameter.
        """
        header = self.headers.get('Content-Type')
        if not header:
            return None
        match = CHARSET_RE.search(header)
        if match:
            return match.group(1)
        return None

    def _charset__set(self, charset):
        if charset is None:
            self._charset__del()
            return
        header = self.headers.get('Content-Type', None)
        if header is None:
            raise AttributeError("You cannot set the charset when no "
                                 "content-type is defined")
        match = CHARSET_RE.search(header)
        if match:
            header = header[:match.start()] + header[match.end():]
        header += '; charset=%s' % charset
        self.headers['Content-Type'] = header

    def _charset__del(self):
        header = self.headers.pop('Content-Type', None)
        if header is None:
            # Don't need to remove anything
            return
        match = CHARSET_RE.search(header)
        if match:
            header = header[:match.start()] + header[match.end():]
        self.headers['Content-Type'] = header

    charset = property(_charset__get, _charset__set, _charset__del,
                       doc=_charset__get.__doc__)

    #
    # content_type
    #

    def _content_type__get(self):
        """
        Get/set the ``Content-Type`` header. If no ``Content-Type`` header is
        set, this will return ``None``.

        .. versionchanged:: 1.7

            Setting a new ``Content-Type`` will remove all ``Content-Type``
            parameters and reset the ``charset`` to the default if the
            ``Content-Type`` is ``text/*`` or XML (``application/xml`` or
            ``*/*+xml``).

            To preserve all ``Content-Type`` parameters, you may use the
            following code:

            .. code-block:: python

                resp = Response()
                params = resp.content_type_params
                resp.content_type = 'application/something'
                resp.content_type_params = params
        """
        header = self.headers.get('Content-Type')
        if not header:
            return None
        return header.split(';', 1)[0]

    def _content_type__set(self, value):
        if not value:
            self._content_type__del()
            return
        else:
            if PY2 and isinstance(value, text_type):
                value = value.encode("latin-1")

            if not isinstance(value, string_types):
                raise TypeError("content_type requires value to be of string_types")

            content_type = value

            # Set up the charset if the content-type doesn't have one

            has_charset = 'charset=' in content_type

            new_charset = None

            if (
                not has_charset and
                self.default_charset
            ):
                new_charset = self.default_charset

            # Optimize for the default_content_type as shipped by
            # WebOb, becuase we know that 'text/html' has a charset,
            # otherwise add a charset if the content_type has a charset.
            #
            # We add the default charset if the content-type is "texty".
            if (
                new_charset and
                (
                    content_type == 'text/html' or
                    _content_type_has_charset(content_type)
                )
            ):
                content_type += '; charset=' + new_charset

            self.headers['Content-Type'] = content_type

    def _content_type__del(self):
        self.headers.pop('Content-Type', None)

    content_type = property(_content_type__get, _content_type__set,
                            _content_type__del, doc=_content_type__get.__doc__)

    #
    # content_type_params
    #

    def _content_type_params__get(self):
        """
        A dictionary of all the parameters in the content type.

        (This is not a view, set to change, modifications of the dict will not
        be applied otherwise.)
        """
        params = self.headers.get('Content-Type', '')
        if ';' not in params:
            return {}
        params = params.split(';', 1)[1]
        result = {}
        for match in _PARAM_RE.finditer(params):
            result[match.group(1)] = match.group(2) or match.group(3) or ''
        return result

    def _content_type_params__set(self, value_dict):
        if not value_dict:
            self._content_type_params__del()
            return

        params = []
        for k, v in sorted(value_dict.items()):
            if not _OK_PARAM_RE.search(v):
                v = '"%s"' % v.replace('"', '\\"')
            params.append('; %s=%s' % (k, v))
        ct = self.headers.pop('Content-Type', '').split(';', 1)[0]
        ct += ''.join(params)
        self.headers['Content-Type'] = ct

    def _content_type_params__del(self):
        self.headers['Content-Type'] = self.headers.get(
            'Content-Type', '').split(';', 1)[0]

    content_type_params = property(
        _content_type_params__get,
        _content_type_params__set,
        _content_type_params__del,
        _content_type_params__get.__doc__
    )

    #
    # set_cookie, unset_cookie, delete_cookie, merge_cookies
    #

    def set_cookie(self, name, value='', max_age=None,
                   path='/', domain=None, secure=False, httponly=False,
                   comment=None, expires=None, overwrite=False,
                   samesite=None):
        """
        Set (add) a cookie for the response.

        Arguments are:

        ``name``

           The cookie name.

        ``value``

           The cookie value, which should be a string or ``None``.  If
           ``value`` is ``None``, it's equivalent to calling the
           :meth:`webob.response.Response.unset_cookie` method for this
           cookie key (it effectively deletes the cookie on the client).

        ``max_age``

           An integer representing a number of seconds, ``datetime.timedelta``,
           or ``None``. This value is used as the ``Max-Age`` of the generated
           cookie.  If ``expires`` is not passed and this value is not
           ``None``, the ``max_age`` value will also influence the ``Expires``
           value of the cookie (``Expires`` will be set to ``now`` +
           ``max_age``).  If this value is ``None``, the cookie will not have a
           ``Max-Age`` value (unless ``expires`` is set). If both ``max_age``
           and ``expires`` are set, this value takes precedence.

        ``path``

           A string representing the cookie ``Path`` value.  It defaults to
           ``/``.

        ``domain``

           A string representing the cookie ``Domain``, or ``None``.  If
           domain is ``None``, no ``Domain`` value will be sent in the
           cookie.

        ``secure``

           A boolean.  If it's ``True``, the ``secure`` flag will be sent in
           the cookie, if it's ``False``, the ``secure`` flag will not be
           sent in the cookie.

        ``httponly``

           A boolean.  If it's ``True``, the ``HttpOnly`` flag will be sent
           in the cookie, if it's ``False``, the ``HttpOnly`` flag will not
           be sent in the cookie.

        ``samesite``

          A string representing the ``SameSite`` attribute of the cookie or
          ``None``. If samesite is ``None`` no ``SameSite`` value will be sent
          in the cookie. Should only be ``"strict"``, ``"lax"``, or ``"none"``.

        ``comment``

           A string representing the cookie ``Comment`` value, or ``None``.
           If ``comment`` is ``None``, no ``Comment`` value will be sent in
           the cookie.

        ``expires``

           A ``datetime.timedelta`` object representing an amount of time,
           ``datetime.datetime`` or ``None``. A non-``None`` value is used to
           generate the ``Expires`` value of the generated cookie. If
           ``max_age`` is not passed, but this value is not ``None``, it will
           influence the ``Max-Age`` header. If this value is ``None``, the
           ``Expires`` cookie value will be unset (unless ``max_age`` is set).
           If ``max_age`` is set, it will be used to generate the ``expires``
           and this value is ignored.

           If a ``datetime.datetime`` is provided it has to either be timezone
           aware or be based on UTC. ``datetime.datetime`` objects that are
           local time are not supported. Timezone aware ``datetime.datetime``
           objects are converted to UTC.

           This argument will be removed in future versions of WebOb (version
           1.9).

        ``overwrite``

           If this key is ``True``, before setting the cookie, unset any
           existing cookie.

        """

        # Remove in WebOb 1.10
        if expires:
            warn_deprecation('Argument "expires" will be removed in a future '
                             'version of WebOb, please use "max_age".', 1.10, 1)

        if overwrite:
            self.unset_cookie(name, strict=False)

        # If expires is set, but not max_age we set max_age to expires
        if not max_age and isinstance(expires, timedelta):
            max_age = expires

        # expires can also be a datetime
        if not max_age and isinstance(expires, datetime):

            # If expires has a timezone attached, convert it to UTC
            if expires.tzinfo and expires.utcoffset():
                expires = (expires - expires.utcoffset()).replace(tzinfo=None)

            max_age = expires - datetime.utcnow()

        value = bytes_(value, 'utf-8')

        cookie = make_cookie(name, value, max_age=max_age, path=path,
                domain=domain, secure=secure, httponly=httponly,
                comment=comment, samesite=samesite)
        self.headerlist.append(('Set-Cookie', cookie))

    def delete_cookie(self, name, path='/', domain=None):
        """
        Delete a cookie from the client.  Note that ``path`` and ``domain``
        must match how the cookie was originally set.

        This sets the cookie to the empty string, and ``max_age=0`` so
        that it should expire immediately.
        """
        self.set_cookie(name, None, path=path, domain=domain)

    def unset_cookie(self, name, strict=True):
        """
        Unset a cookie with the given name (remove it from the response).
        """
        existing = self.headers.getall('Set-Cookie')
        if not existing and not strict:
            return
        cookies = Cookie()
        for header in existing:
            cookies.load(header)
        if isinstance(name, text_type):
            name = name.encode('utf8')
        if name in cookies:
            del cookies[name]
            del self.headers['Set-Cookie']
            for m in cookies.values():
                self.headerlist.append(('Set-Cookie', m.serialize()))
        elif strict:
            raise KeyError("No cookie has been set with the name %r" % name)

    def merge_cookies(self, resp):
        """Merge the cookies that were set on this response with the
        given ``resp`` object (which can be any WSGI application).

        If the ``resp`` is a :class:`webob.Response` object, then the
        other object will be modified in-place.
        """
        if not self.headers.get('Set-Cookie'):
            return resp
        if isinstance(resp, Response):
            for header in self.headers.getall('Set-Cookie'):
                resp.headers.add('Set-Cookie', header)
            return resp
        else:
            c_headers = [h for h in self.headerlist if
                         h[0].lower() == 'set-cookie']
            def repl_app(environ, start_response):
                def repl_start_response(status, headers, exc_info=None):
                    return start_response(status, headers + c_headers,
                                          exc_info=exc_info)
                return resp(environ, repl_start_response)
            return repl_app

    #
    # cache_control
    #

    _cache_control_obj = None

    def _cache_control__get(self):
        """
        Get/set/modify the Cache-Control header (`HTTP spec section 14.9
        <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.9>`_).
        """
        value = self.headers.get('cache-control', '')
        if self._cache_control_obj is None:
            self._cache_control_obj = CacheControl.parse(
                value, updates_to=self._update_cache_control, type='response')
            self._cache_control_obj.header_value = value
        if self._cache_control_obj.header_value != value:
            new_obj = CacheControl.parse(value, type='response')
            self._cache_control_obj.properties.clear()
            self._cache_control_obj.properties.update(new_obj.properties)
            self._cache_control_obj.header_value = value
        return self._cache_control_obj

    def _cache_control__set(self, value):
        # This actually becomes a copy
        if not value:
            value = ""
        if isinstance(value, dict):
            value = CacheControl(value, 'response')
        if isinstance(value, text_type):
            value = str(value)
        if isinstance(value, str):
            if self._cache_control_obj is None:
                self.headers['Cache-Control'] = value
                return
            value = CacheControl.parse(value, 'response')
        cache = self.cache_control
        cache.properties.clear()
        cache.properties.update(value.properties)

    def _cache_control__del(self):
        self.cache_control = {}

    def _update_cache_control(self, prop_dict):
        value = serialize_cache_control(prop_dict)
        if not value:
            if 'Cache-Control' in self.headers:
                del self.headers['Cache-Control']
        else:
            self.headers['Cache-Control'] = value

    cache_control = property(
        _cache_control__get, _cache_control__set,
        _cache_control__del, doc=_cache_control__get.__doc__)

    #
    # cache_expires
    #

    def _cache_expires(self, seconds=0, **kw):
        """
            Set expiration on this request.  This sets the response to
            expire in the given seconds, and any other attributes are used
            for ``cache_control`` (e.g., ``private=True``).
        """
        if seconds is True:
            seconds = 0
        elif isinstance(seconds, timedelta):
            seconds = timedelta_to_seconds(seconds)
        cache_control = self.cache_control
        if seconds is None:
            pass
        elif not seconds:
            # To really expire something, you have to force a
            # bunch of these cache control attributes, and IE may
            # not pay attention to those still so we also set
            # Expires.
            cache_control.no_store = True
            cache_control.no_cache = True
            cache_control.must_revalidate = True
            cache_control.max_age = 0
            cache_control.post_check = 0
            cache_control.pre_check = 0
            self.expires = datetime.utcnow()
            if 'last-modified' not in self.headers:
                self.last_modified = datetime.utcnow()
            self.pragma = 'no-cache'
        else:
            cache_control.properties.clear()
            cache_control.max_age = seconds
            self.expires = datetime.utcnow() + timedelta(seconds=seconds)
            self.pragma = None
        for name, value in kw.items():
            setattr(cache_control, name, value)

    cache_expires = property(lambda self: self._cache_expires, _cache_expires)

    #
    # encode_content, decode_content, md5_etag
    #

    def encode_content(self, encoding='gzip', lazy=False):
        """
        Encode the content with the given encoding (only ``gzip`` and
        ``identity`` are supported).
        """
        assert encoding in ('identity', 'gzip'), \
            "Unknown encoding: %r" % encoding
        if encoding == 'identity':
            self.decode_content()
            return
        if self.content_encoding == 'gzip':
            return
        if lazy:
            self.app_iter = gzip_app_iter(self._app_iter)
            self.content_length = None
        else:
            self.app_iter = list(gzip_app_iter(self._app_iter))
            self.content_length = sum(map(len, self._app_iter))
        self.content_encoding = 'gzip'

    def decode_content(self):
        content_encoding = self.content_encoding or 'identity'
        if content_encoding == 'identity':
            return
        if content_encoding not in ('gzip', 'deflate'):
            raise ValueError(
                "I don't know how to decode the content %s" % content_encoding)
        if content_encoding == 'gzip':
            from gzip import GzipFile
            from io import BytesIO
            gzip_f = GzipFile(filename='', mode='r', fileobj=BytesIO(self.body))
            self.body = gzip_f.read()
            self.content_encoding = None
            gzip_f.close()
        else:
            # Weird feature: http://bugs.python.org/issue5784
            self.body = zlib.decompress(self.body, -15)
            self.content_encoding = None

    def md5_etag(self, body=None, set_content_md5=False):
        """
        Generate an etag for the response object using an MD5 hash of
        the body (the ``body`` parameter, or ``self.body`` if not given).

        Sets ``self.etag``.

        If ``set_content_md5`` is ``True``, sets ``self.content_md5`` as well.
        """
        if body is None:
            body = self.body
        md5_digest = md5(body).digest()
        md5_digest = b64encode(md5_digest)
        md5_digest = md5_digest.replace(b'\n', b'')
        md5_digest = native_(md5_digest)
        self.etag = md5_digest.strip('=')
        if set_content_md5:
            self.content_md5 = md5_digest

    @staticmethod
    def _make_location_absolute(environ, value):
        if SCHEME_RE.search(value):
            return value

        new_location = urlparse.urljoin(_request_uri(environ), value)
        return new_location

    def _abs_headerlist(self, environ):
        # Build the headerlist, if we have a Location header, make it absolute
        return [
            (k, v) if k.lower() != 'location'
            else (k, self._make_location_absolute(environ, v))
            for (k, v)
            in self._headerlist
        ]

    #
    # __call__, conditional_response_app
    #

    def __call__(self, environ, start_response):
        """
        WSGI application interface
        """
        if self.conditional_response:
            return self.conditional_response_app(environ, start_response)

        headerlist = self._abs_headerlist(environ)

        start_response(self.status, headerlist)
        if environ['REQUEST_METHOD'] == 'HEAD':
            # Special case here...
            return EmptyResponse(self._app_iter)
        return self._app_iter

    _safe_methods = ('GET', 'HEAD')

    def conditional_response_app(self, environ, start_response):
        """
        Like the normal ``__call__`` interface, but checks conditional headers:

            * ``If-Modified-Since``   (``304 Not Modified``; only on ``GET``,
              ``HEAD``)
            * ``If-None-Match``       (``304 Not Modified``; only on ``GET``,
              ``HEAD``)
            * ``Range``               (``406 Partial Content``; only on ``GET``,
              ``HEAD``)
        """
        req = BaseRequest(environ)

        headerlist = self._abs_headerlist(environ)

        method = environ.get('REQUEST_METHOD', 'GET')
        if method in self._safe_methods:
            status304 = False
            if req.if_none_match and self.etag:
                status304 = self.etag in req.if_none_match
            elif req.if_modified_since and self.last_modified:
                status304 = self.last_modified <= req.if_modified_since
            if status304:
                start_response('304 Not Modified', filter_headers(headerlist))
                return EmptyResponse(self._app_iter)
        if (
            req.range and self in req.if_range and
            self.content_range is None and
            method in ('HEAD', 'GET') and
            self.status_code == 200 and
            self.content_length is not None
        ):
            content_range = req.range.content_range(self.content_length)
            if content_range is None:
                iter_close(self._app_iter)
                body = bytes_("Requested range not satisfiable: %s" % req.range)
                headerlist = [
                    ('Content-Length', str(len(body))),
                    ('Content-Range', str(ContentRange(None, None,
                                                       self.content_length))),
                    ('Content-Type', 'text/plain'),
                ] + filter_headers(headerlist)
                start_response('416 Requested Range Not Satisfiable',
                               headerlist)
                if method == 'HEAD':
                    return ()
                return [body]
            else:
                app_iter = self.app_iter_range(content_range.start,
                                               content_range.stop)
                if app_iter is not None:
                    # the following should be guaranteed by
                    # Range.range_for_length(length)
                    assert content_range.start is not None
                    headerlist = [
                        ('Content-Length',
                         str(content_range.stop - content_range.start)),
                        ('Content-Range', str(content_range)),
                    ] + filter_headers(headerlist, ('content-length',))
                    start_response('206 Partial Content', headerlist)
                    if method == 'HEAD':
                        return EmptyResponse(app_iter)
                    return app_iter

        start_response(self.status, headerlist)
        if method == 'HEAD':
            return EmptyResponse(self._app_iter)
        return self._app_iter

    def app_iter_range(self, start, stop):
        """
        Return a new ``app_iter`` built from the response ``app_iter``, that
        serves up only the given ``start:stop`` range.
        """
        app_iter = self._app_iter
        if hasattr(app_iter, 'app_iter_range'):
            return app_iter.app_iter_range(start, stop)
        return AppIterRange(app_iter, start, stop)


def filter_headers(hlist, remove_headers=('content-length', 'content-type')):
    return [h for h in hlist if (h[0].lower() not in remove_headers)]


def iter_file(file, block_size=1 << 18): # 256Kb
    while True:
        data = file.read(block_size)
        if not data:
            break
        yield data

class ResponseBodyFile(object):
    mode = 'wb'
    closed = False

    def __init__(self, response):
        """
        Represents a :class:`~Response` as a file like object.
        """
        self.response = response
        self.write = response.write

    def __repr__(self):
        return '<body_file for %r>' % self.response

    encoding = property(
        lambda self: self.response.charset,
        doc="The encoding of the file (inherited from response.charset)"
    )

    def writelines(self, seq):
        """
        Write a sequence of lines to the response.
        """
        for item in seq:
            self.write(item)

    def close(self):
        raise NotImplementedError("Response bodies cannot be closed")

    def flush(self):
        pass

    def tell(self):
        """
        Provide the current location where we are going to start writing.
        """
        if not self.response.has_body:
            return 0

        return sum([len(chunk) for chunk in self.response.app_iter])


class AppIterRange(object):
    """
    Wraps an ``app_iter``, returning just a range of bytes.
    """

    def __init__(self, app_iter, start, stop):
        assert start >= 0, "Bad start: %r" % start
        assert stop is None or (stop >= 0 and stop >= start), (
            "Bad stop: %r" % stop)
        self.app_iter = iter(app_iter)
        self._pos = 0 # position in app_iter
        self.start = start
        self.stop = stop

    def __iter__(self):
        return self

    def _skip_start(self):
        start, stop = self.start, self.stop
        for chunk in self.app_iter:
            self._pos += len(chunk)
            if self._pos < start:
                continue
            elif self._pos == start:
                return b''
            else:
                chunk = chunk[start - self._pos:]
                if stop is not None and self._pos > stop:
                    chunk = chunk[:stop - self._pos]
                    assert len(chunk) == stop - start
                return chunk
        else:
            raise StopIteration()

    def next(self):
        if self._pos < self.start:
            # need to skip some leading bytes
            return self._skip_start()
        stop = self.stop
        if stop is not None and self._pos >= stop:
            raise StopIteration

        chunk = next(self.app_iter)
        self._pos += len(chunk)

        if stop is None or self._pos <= stop:
            return chunk
        else:
            return chunk[:stop - self._pos]

    __next__ = next # py3

    def close(self):
        iter_close(self.app_iter)


class EmptyResponse(object):
    """
    An empty WSGI response.

    An iterator that immediately stops. Optionally provides a close
    method to close an underlying ``app_iter`` it replaces.
    """

    def __init__(self, app_iter=None):
        if app_iter is not None and hasattr(app_iter, 'close'):
            self.close = app_iter.close

    def __iter__(self):
        return self

    def __len__(self):
        return 0

    def next(self):
        raise StopIteration()

    __next__ = next # py3

def _is_xml(content_type):
    return (
        content_type.startswith('application/xml') or
        (
            content_type.startswith('application/') and
            content_type.endswith('+xml')
        ) or
        (
            content_type.startswith('image/') and
            content_type.endswith('+xml')
        )
    )

def _content_type_has_charset(content_type):
    return (
        content_type.startswith('text/') or
        _is_xml(content_type)
    )

def _request_uri(environ):
    """Like ``wsgiref.url.request_uri``, except eliminates ``:80`` ports.

    Returns the full request URI."""
    url = environ['wsgi.url_scheme'] + '://'

    if environ.get('HTTP_HOST'):
        url += environ['HTTP_HOST']
    else:
        url += environ['SERVER_NAME'] + ':' + environ['SERVER_PORT']
    if url.endswith(':80') and environ['wsgi.url_scheme'] == 'http':
        url = url[:-3]
    elif url.endswith(':443') and environ['wsgi.url_scheme'] == 'https':
        url = url[:-4]

    if PY2:
        script_name = environ.get('SCRIPT_NAME', '/')
        path_info = environ.get('PATH_INFO', '')
    else:
        script_name = bytes_(environ.get('SCRIPT_NAME', '/'), 'latin-1')
        path_info = bytes_(environ.get('PATH_INFO', ''), 'latin-1')

    url += url_quote(script_name)
    qpath_info = url_quote(path_info)
    if 'SCRIPT_NAME' not in environ:
        url += qpath_info[1:]
    else:
        url += qpath_info
    return url


def iter_close(iter):
    if hasattr(iter, 'close'):
        iter.close()

def gzip_app_iter(app_iter):
    size = 0
    crc = zlib.crc32(b"") & 0xffffffff
    compress = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS,
                                zlib.DEF_MEM_LEVEL, 0)

    yield _gzip_header
    for item in app_iter:
        size += len(item)
        crc = zlib.crc32(item, crc) & 0xffffffff

        # The compress function may return zero length bytes if the input is
        # small enough; it buffers the input for the next iteration or for a
        # flush.
        result = compress.compress(item)
        if result:
            yield result

    # Similarly, flush may also not yield a value.
    result = compress.flush()
    if result:
        yield result
    yield struct.pack("<2L", crc, size & 0xffffffff)

def _error_unicode_in_app_iter(app_iter, body):
    app_iter_repr = repr(app_iter)
    if len(app_iter_repr) > 50:
        app_iter_repr = (
            app_iter_repr[:30] + '...' + app_iter_repr[-10:])
    raise TypeError(
        'An item of the app_iter (%s) was text, causing a '
        'text body: %r' % (app_iter_repr, body))



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent import scope
from contrast.agent.assess.adjusted_span import AdjustedSpan
from .base import StringToken, FormatToken as BaseFormatToken


# This regex was built from the following description:
# https://docs.python.org/3/library/stdtypes.html#printf-style-string-formatting
from contrast.utils.string_utils import ensure_string, ensure_binary

FORMAT_REGEX_STRING = r"""
    %                           # required for all format strings
    (?:\(([^\d\W]\w*)\))?       # optional mapping key (identifer name)
    ([#0\- +])?                 # optional conversion flags
    ([1-9]+[0-9]?|\*)?          # optional minimum field width
    (?:\.([0-9]+|\*))?          # optional precision
    ([hlL])?                    # optional length modifier (ignored)
    ([diouxXeEfFgGcrsa%])       # conversion type (required)
    """

UNICODE_REGEX_FLAGS = re.X | re.UNICODE
FORMAT_REGEX = re.compile(FORMAT_REGEX_STRING, UNICODE_REGEX_FLAGS)
# For matching bytes and bytearray objects
FORMAT_REGEX_BYTES = re.compile(FORMAT_REGEX_STRING.encode(), re.X)


class BaseCFormatToken(BaseFormatToken):
    def __init__(self, match, index):
        self.start, self.end = match.span()
        (
            self.mapping,
            self.conversion,
            self.width,
            self.precision,
            _,
            self.type,
        ) = match.groups()
        self.match = match

        # From the Python documentation: if either the width or the precision
        # is specified as an '*' (asterisk), the actual width or precision is
        # read from the next element of the argument tuple, and the value to
        # convert comes after the width and/or precision.
        if self.width == self.wildcard:
            index += 1
        if self.precision == self.wildcard:
            index += 1

        self.index = index

    def _convert_int(self, val):
        return self.encode(str(int(val))) if val else self.empty

    def _build_width_and_precision(self, args):
        if self.width == self.wildcard and self.precision == self.wildcard:
            width, precision = args[self.index - 2 : self.index]
        elif self.width == self.wildcard:
            width, precision = args[self.index - 1], self.precision
        elif self.precision == self.wildcard:
            width, precision = self.width, args[self.index - 1]
        else:
            width, precision = self.width, self.precision

        width = self._convert_int(width)
        precision = self.dot + self._convert_int(precision) if precision else self.empty

        return self.formatter % (width, precision)

    def format(self, args, kwargs):
        string = self.get_arg(args, kwargs)

        fmt = self.percent
        if self.conversion:
            fmt += self.conversion
        fmt += self._build_width_and_precision(args)
        fmt += self.type

        return string, fmt % (string,)

    def get_arg(self, args, kwargs):
        if self.type == self.percent:
            return self.percent
        if self.mapping:
            return args[0][self.mapping]
        return args[self.index]

    @property
    def next_index(self):
        # This ensures that we handle indexing properly in cases where there
        # are escape sequences in the string that is being tokenized.
        # We don't increment the index in these cases since the token doesn't
        # correspond to any of the input arguments to be formatted.
        return self.index if self.type == self.percent else self.index + 1

    @property
    def span(self):
        return AdjustedSpan(self.start, self.end)

    def __len__(self):
        return self.end - self.start

    def __repr__(self):
        reprstr = "<{}({})>"
        clsname = self.__class__.__name__
        return reprstr.format(clsname, self.match)


def _build_token_class(name, encoder):
    cls = type(name, (BaseCFormatToken,), {})

    consts = dict(wildcard="*", empty="", dot=".", percent="%", formatter="%s%s")
    for name, val in consts.items():
        setattr(cls, name, encoder(val))

    def encode(self, val):
        return encoder(val)

    setattr(cls, "encode", encode)

    return cls


StrFormatToken = _build_token_class("StrFormatToken", ensure_string)
BytesFormatToken = _build_token_class("BytesFormatToken", ensure_binary)


def tokenize_format(format_string):
    tokens = []
    token_index = 0

    is_string = isinstance(format_string, str)
    token_cls = StrFormatToken if is_string else BytesFormatToken

    regex = FORMAT_REGEX if is_string else FORMAT_REGEX_BYTES

    start_index = 0
    with scope.contrast_scope():
        # contrast scope required here to prevent analysis in finditer patch.
        for match in regex.finditer(format_string):
            start, end = match.span()
            if start_index != start:
                string = format_string[start_index:start]
                tokens.append(StringToken(string, start_index))

            fmt_token = token_cls(match, token_index)
            tokens.append(fmt_token)

            start_index = end
            token_index = fmt_token.next_index

    if start_index < len(format_string):
        string = format_string[start_index:]
        tokens.append(StringToken(string, start_index))

    return tokens



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
import platform
import tempfile
import time
from glob import glob
from io import open
from os import environ, path, system
from shutil import rmtree
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.install_lib import install_lib
import warnings

root_dir = path.abspath(path.dirname(__file__))


AGENT_LIB_REQUIREMENT = "~=0.5.4"


def read(*parts):
    with open(path.join(root_dir, *parts), encoding="utf-8") as f:
        return f.read()


def is_arm():
    machine = platform.machine()
    return "arm" in machine or "aarch" in machine


version_specifier = sys.version_info[:2]


extension_path = path.join("contrast", "assess_extensions")
extension_source_dir = path.join("src", extension_path)
version_dir = path.join(extension_source_dir, "py{}{}".format(*version_specifier))
common_dir = path.join(extension_source_dir, "common")

if sys.platform.startswith("darwin"):
    link_args = ["-rpath", "@loader_path"]
    platform_args = []
else:
    platform_args = ["-Wno-cast-function-type"]
    link_args = []

if version_specifier == (3, 7):
    warnings.warn(
        "Support for Python 3.7 is deprecated and will be removed in a future release of the Contrast Agent",
        DeprecationWarning,
    )

debug = environ.get("ASSESS_DEBUG")
debug_args = ["-g", "-O1"] if debug else []
macros = [("ASSESS_DEBUG", 1)] if debug else []
macros.append(("EXTENSION_BUILD_TIME", '"{}"'.format(time.ctime())))

strict_build_args = ["-Werror"] if environ.get("CONTRAST_STRICT_BUILD") else []

NO_FUNCHOOK = environ.get("CONTRAST_NO_FUNCHOOK") or is_arm()
if NO_FUNCHOOK:
    macros.append(("NO_FUNCHOOK", 1))
    c_sources = [
        path.join(common_dir, name)
        for name in [
            "patches.c",
            "scope.c",
            "logging.c",
            "intern.c",
            "propagate.c",
            "format.c",
            "repr.c",
            "repeat.c",
            "streams.c",
            "subscript.c",
            "cast.c",
            "trace.c",
        ]
    ]
    c_sources.extend(glob(path.join(extension_source_dir, "py3", "patches.c")))
    c_sources.extend(
        [
            path.join(version_dir, name + ".c")
            for name in ["iobase", "bytesio", "stringio"]
        ]
    )
    libraries = []
else:
    c_sources = glob(path.join(common_dir, "*.c")) + glob(path.join(version_dir, "*.c"))
    # Add source files common to all python3 versions
    c_sources.extend(glob(path.join(extension_source_dir, "py3", "*.c")))
    libraries = ["funchook"]


extensions = [
    Extension(
        "contrast.assess_extensions.cs_str",
        c_sources,
        libraries=libraries,
        include_dirs=[
            extension_source_dir,
            path.join(extension_source_dir, "include"),
        ],
        library_dirs=[extension_source_dir],
        # Path relative to the .so itself (works for gnu-ld)
        runtime_library_dirs=["$ORIGIN"],
        extra_compile_args=[
            "-Wall",
            "-Wextra",
            "-Wno-unused-parameter",
            "-Wmissing-field-initializers",
        ]
        + strict_build_args
        + debug_args
        + platform_args,
        extra_link_args=link_args,
        define_macros=macros,
    )
]

tempdir = None
funchook_temp = None

build_err_msg = "Failed to build Contrast C extension.\n"

autotools_err_msg = """
It is necessary for autotools (autoconf, automake) to be installed in order for
Contrast to build properly. On lightweight systems such as Alpine, it may be
necessary to install linux-headers if they are not available already. Some
other systems may require "build essential" packages to be installed.
"""


class ContrastBuildExt(build_ext):
    def run(self):
        if NO_FUNCHOOK:
            build_ext.run(self)
            return

        if system("/bin/sh src/contrast/assess_extensions/build_funchook.sh") != 0:
            raise RuntimeError(f"{build_err_msg}\n{autotools_err_msg}")

        build_ext.run(self)

        global tempdir
        global funchook_temp

        ext = "dylib" if sys.platform.startswith("darwin") else "so"
        funchook_name = "libfunchook.{}".format(ext)
        funchook = path.join(extension_source_dir, funchook_name)

        tempdir = tempfile.mkdtemp("contrast-build")
        funchook_temp = path.join(tempdir, funchook_name)
        self.copy_file(funchook, funchook_temp)


class ContrastInstallLib(install_lib):
    def run(self):
        install_lib.run(self)

        if NO_FUNCHOOK:
            return

        if funchook_temp is not None:
            dest_dir = path.join(self.install_dir, extension_path)
            self.copy_file(funchook_temp, dest_dir)
            rmtree(tempdir)


setup(
    install_requires=[
        "requests>=2.4.2,<3",
        f"contrast-agent-lib{AGENT_LIB_REQUIREMENT}",
    ],
    cmdclass=dict(build_ext=ContrastBuildExt, install_lib=ContrastInstallLib),
    ext_modules=extensions,
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.source_policy import get_parent_ids
from contrast.agent.assess.tag import Tag
from contrast.utils.assess import tracking_util
from contrast.utils.assess.tag_utils import combine, is_covered
from contrast.utils.decorators import cached_property, fail_quietly
from contrast.agent.assess.utils import (
    is_tracked,
    get_properties,
    track_string,
)

SUPPORTED_TYPES = (
    str,
    bytes,
    bytearray,
)


class BasePropagator(object):
    def __init__(self, node, preshift, target):
        """
        :param node: instance of PropagationNode
        """
        self.node = node
        self.preshift = preshift
        self.target = target
        self.target_properties = None

    @cached_property
    def sources(self):
        """
        Get all the sources for the propagation node.
        """
        return self.node.get_matching_sources(self.preshift)

    @property
    def any_source_tracked(self):
        return any(tracking_util.recursive_is_tracked(s) for s in self.sources)

    @property
    def needs_propagation(self):
        if not isinstance(self.target, SUPPORTED_TYPES):
            return False

        if tracking_util.recursive_is_tracked(self.target):
            return True

        if not self.preshift:
            return False

        return self.inputs_require_propagation

    @property
    def inputs_require_propagation(self):
        return self.any_source_tracked

    def track_target(self):
        for source in self.sources:
            if is_tracked(source):
                track_string(self.target)

    def propagate(self):
        """
        Any propagators that inherit from BasePropagator and are able to follow
        this same propagation pattern should define _propagate.

        Some propagators continue to define their own propagate method.
        """
        self.target_properties = get_properties(self.target)
        if self.target_properties is None:
            return

        self.first_source = self.sources[0]
        self._propagate()
        self.target_properties.cleanup_tags()

    def get_parent_ids(self, ret):
        """Some derived classes may need to override this method"""
        return get_parent_ids(
            self.node, self.preshift.obj, ret, self.preshift.args, self.preshift.kwargs
        )

    def build_event(self, target_properties, ret):
        parent_ids = self.get_parent_ids(ret)
        target_properties.build_event(
            self.node,
            self.target,
            self.preshift.obj,
            ret,
            self.preshift.args,
            self.preshift.kwargs,
            parent_ids,
            None,
        )

    def splat_tags(self, source, ret):
        """
        SPLAT copies tags for cases where the string transformation is unknown

        SPLAT is subject to the following rules:
            1. Any UNTRUSTED tags on the input are applied to the entire length of the
               output.
            2. Any other tags that cover the entire length of the input are copied to
               the entire length of the output.
            3. Any tags that cover all UNTRUSTED ranges on the input are copied to the
               entire length of the output.
            4. No other tags are copied.

        These rules are intended to minimize the number of false negatives that result
        from accidentally copying safe tags from the input to regions of the output
        that are not actually safe.
        """
        source_props = get_properties(source)
        if source_props is None:
            return

        ret_props = get_properties(ret)
        if ret_props is None:
            return

        ret_length = len(ret)

        untrusted_ranges = source_props.tags.get("UNTRUSTED") or []
        if untrusted_ranges:
            # splat UNTRUSTED ranges across the entire output
            ret_props.add_tag("UNTRUSTED", AdjustedSpan(0, ret_length))

        for name, ranges in source_props.tags.items():
            if name == "UNTRUSTED":
                continue

            if untrusted_ranges and is_covered(untrusted_ranges, ranges):
                ret_props.add_tag(name, AdjustedSpan(0, ret_length))

            combined = combine(ranges)
            if combined[0] == Tag(len(source), 0):
                ret_props.add_tag(name, AdjustedSpan(0, ret_length))

    def apply_tags(self, node, target, span=None):
        target_properties = get_properties(target)
        if not target_properties:
            return

        span = span or AdjustedSpan(0, len(target))
        for tag in node.tags:
            target_properties.add_tag(tag, span)

    def apply_untags(self, node, target):
        target_properties = get_properties(target)

        if target_properties:
            for tag in node.untags:
                target_properties.delete_tag(tag)

    def add_tags(self):
        if self.node.tags:
            self.apply_tags(self.node, self.target)

        if self.node.untags:
            self.apply_untags(self.node, self.target)

    def add_tags_and_properties(self, ret):
        self.add_tags()

        target_properties = get_properties(self.target)

        # ignore un-tracked targets; possible interning
        if target_properties is None:
            return

        target_properties.add_properties(self.node.properties)

        self.build_event(target_properties, ret)

    def reset_tags(self):
        """
        Some propagators will store the original set of tags but then restore them.
        This is because PropagationNode are only initialized once (for the entire server-startup cycle)
        from their policy.json representation, so if a propagator has to add/delete a tag, we need
        to restore the original tags.

        In the future, we could avoid doing this reset by re-instantiating nodes
        every time, albeit with a cost to performance.
        """
        pass

    @fail_quietly("Error during propagation")
    def track_and_propagate(self, ret):
        self.track_target()
        self.propagate()
        self.add_tags_and_properties(ret)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from enum import Enum, auto

from contrast.api import TypeCheckedProperty


class DocumentType(Enum):
    NORMAL = auto()
    JSON = auto()
    XML = auto()


# This is an exact representation of TS InputType. Not all are currently used, since we
# have to map from agent-lib's own input type to this enum.
class InputType(Enum):
    COOKIE_NAME = auto()
    COOKIE_VALUE = auto()
    HEADER = auto()
    PARAMETER_NAME = auto()
    PARAMETER_VALUE = auto()
    QUERYSTRING = auto()
    URI = auto()
    SOCKET = auto()
    JSON_VALUE = auto()
    JSON_ARRAYED_VALUE = auto()
    MULTIPART_CONTENT_TYPE = auto()
    MULTIPART_VALUE = auto()
    MULTIPART_FIELD_NAME = auto()
    MULTIPART_NAME = auto()
    XML_VALUE = auto()
    DWR_VALUE = auto()
    UNKNOWN = auto()
    METHOD = auto()
    REQUEST = auto()
    URL_PARAMETER = auto()


class UserInput:
    input_type = TypeCheckedProperty(InputType, constructor_arg=InputType.UNKNOWN)
    key = TypeCheckedProperty(str)
    value = TypeCheckedProperty(str)
    path = TypeCheckedProperty(str)
    matcher_ids = TypeCheckedProperty(list)
    document_type = TypeCheckedProperty(
        DocumentType, constructor_arg=DocumentType.NORMAL
    )

    def __init__(
        self, input_type, key, value, path="", matcher_ids=None, document_type=None
    ):
        self.input_type = input_type
        self.key = key
        self.value = value
        self.path = path
        self.matcher_ids = [] if matcher_ids is None else matcher_ids

        self.document_type = (
            DocumentType.NORMAL if document_type is None else document_type
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from pip._vendor import pkg_resources
from contrast.agent import scope
from contrast.utils.namespace import Namespace  # noqa: provides a cleaner import
from contrast.utils.profiler import Profiler  # noqa: provides a cleaner import


def get_installed_distributions():
    """
    Wrapper used to get list of installed distributions in current active environment.
    """
    with scope.contrast_scope():
        return [
            d
            for d in pkg_resources.working_set  # pylint: # pylint: disable=not-an-iterable
        ]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ApplicationStartup(BaseTsAppMessage):
    def __init__(self):
        super().__init__()

        self.body = {
            "instrumentation": {
                "protect": {"enable": self.settings.is_protect_enabled()}
            }
        }

        if self.settings.config.get_session_metadata():
            self.body.update(
                {"session_metadata": self.settings.config.get_session_metadata()}
            )

        if self.settings.config.app_metadata:
            self.body.update({"metadata": self.settings.config.app_metadata})

        if self.settings.config.app_code:
            self.body.update({"code": self.settings.config.app_code})

        if self.settings.config.session_id:
            self.body.update({"session_id": self.settings.config.session_id})

        if self.settings.config.app_tags:
            self.body.update({"tags": self.settings.config.app_tags})

        if self.settings.config.app_group:
            self.body.update({"group": self.settings.config.app_group})

    @property
    def name(self):
        return "application-startup"

    @property
    def path(self):
        return "applications/create"

    @property
    def expected_response_codes(self):
        return [200]

    @property
    def disable_agent_on_401_and_408(self):
        return True

    @property
    def request_method(self):
        return requests.put

    @fail_loudly("Failed to process ApplicationStartup response")
    def process_response(self, response, reporting_client):
        if not self.process_response_code(response, reporting_client):
            return

        body = response.json()

        self.settings.apply_ts_app_settings(body)
        self.settings.process_ts_reactions(body)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

# Dataflow rules
from .cmd_injection import *
from .nosql_injection import *
from .path_traversal import *
from .sql_injection import *
from .redos import *
from .reflected_xss import *
from .ssrf import *
from .trust_boundary_violation import *
from .unsafe_code_execution import *
from .untrusted_deserialization import *
from .unvalidated_redirect import *
from .xxe import *
from .xpath_injection import *
from .prompt_injection import *

# Non-dataflow rules
# NOTE: some of our unit tests are sensitive to the ordering of these imports
from .httponly import *
from .session_timeout import *
from .secure_flag_missing import *
from .session_rewriting import *

# Weird rules
from .crypto import *
